from typing import Callable, List

import numpy as np

__all__ = ["Operator", "ColumnwiseOperator", "SequentialOperator"]


class Operator:
    r"""Class that stores a transformation to be applied to a full dataset or to one column."""

    def __init__(self, fun: Callable[[np.ndarray], np.ndarray]):
        r"""

        Parameters
        ----------
        fun : Callable[[np.ndarray], np.ndarray]
            any function that returns numpy.arrays with the same size as its input.
        """
        self.fun = fun

    def __call__(self, t: np.ndarray):
        return self.fun(t)

    def __str__(self) -> str:
        return f"Operator: {self.fun.__name__}"


class ColumnwiseOperator(Operator):
    """Class that stores a list of operators -- one per column of the dataset to be considered. It defines a pre-processing on the dataset."""

    def __init__(self, ops: List[Callable[[np.ndarray], np.ndarray]]):
        r"""

        Parameters
        ----------
        ops : List[Callable[[np.ndarray], np.ndarray]]
            list of operators or composition of operators, with one per column of the dataset to be pre-processed.
        """
        self.n_cols = len(ops)
        self.ops = ops

    def __call__(self, t: np.ndarray):
        return np.vstack([self.ops[k](t[:, k]) for k in range(self.n_cols)]).T

    def __str__(self) -> str:
        ops_str = [op.__name__ for op in self.ops]
        return f"ColumnwiseOperator: {ops_str}"


class SequentialOperator(Operator):
    r"""Defines an operator on one column as a composition of multiple operators."""

    def __init__(self, ops: List[Operator]):
        r"""

        Parameters
        ----------
        ops : List[Operator]
            the list of operators to be composed, in the order of application.
        """
        self.ops = ops

    def __call__(self, t: np.ndarray):
        for op in self.ops:
            t = op(t)
        return t

    def __str__(self) -> str:
        ops_str = [str(op) for op in self.ops]
        return f"SequentialOperator: {ops_str}"



from .neural_network import *
from .fully_connected import *
from .densely_connected import *
from .polynomial_network import *
from .merging_network import *
from .embedding_network import *

__all__ = []

__all__.extend(neural_network.__all__)
__all__.extend(fully_connected.__all__)
__all__.extend(densely_connected.__all__)
__all__.extend(polynomial_network.__all__)
__all__.extend(merging_network.__all__)
__all__.extend(embedding_network.__all__)



import datetime
import random
from typing import Callable, Dict, List, Literal, Optional, Sequence, Tuple, Union

import numpy as np
import torch
from torch.optim import Optimizer
from torch.optim.lr_scheduler import ConstantLR, ReduceLROnPlateau, _LRScheduler
from torch.utils.data import DataLoader, TensorDataset
from tqdm import tqdm

from nnbma.dataset import MaskDataset, MaskSubset, RegressionDataset, RegressionSubset
from nnbma.networks import NeuralNetwork

from .batch_scheduler import BatchScheduler
from .loss_functions import MaskedLossFunction, MaskOverlay

__all__ = [
    "LearningParameters",
    "learning_procedure",
]


class LearningParameters:
    r"""Specifies the main parameters training, including the loss function to minimize and the stochastic gradient descent strategy."""

    def __init__(
        self,
        loss_fun: Union[
            Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
            MaskedLossFunction,
        ],
        epochs: int,
        batch_size: Union[int, BatchScheduler, None],
        optimizer: Optimizer,
        scheduler: Optional[_LRScheduler] = None,
    ):
        r"""

        Parameters
        ----------
        loss_fun : Union[ Callable[[torch.Tensor, torch.Tensor], torch.Tensor], MaskedLossFunction, ]
            loss function.
        epochs : int
            total number of epochs to perform.
        batch_size : Union[int, BatchScheduler, None]
            batch size value or scheduler to use during training.
        optimizer : Optimizer
            optimizer to use for training.
        scheduler : Optional[_LRScheduler], optional
            learning rate scheduler, by default None
        """
        self.loss_fun = loss_fun
        self.epochs = epochs
        self.batch_size = batch_size
        self.optimizer = optimizer
        if scheduler is None:
            self.scheduler = ConstantLR(optimizer, 1.0)
        else:
            self.scheduler = scheduler

    def __str__(self):
        s = "Learning parameters:\n"
        s += f"\tLoss function: {self.loss_fun}\n"
        s += f"\tEpochs: {self.epochs}\n"
        s += f"\tBatch size: {self.batch_size}\n"
        s += f"\tOptimizer: {self.optimizer}\n"
        s += f"\tScheduler: {self.scheduler}"
        return s


def learning_procedure(
    model: NeuralNetwork,
    dataset: Union[RegressionDataset, Tuple[RegressionDataset, RegressionDataset]],
    learning_parameters: Union[LearningParameters, List[LearningParameters]],
    mask_dataset: Union[
        MaskDataset, Tuple[MaskDataset, MaskDataset], None, Tuple[None, None]
    ] = None,
    train_samples: Optional[Sequence] = None,
    val_samples: Optional[Sequence] = None,
    val_frac: Optional[float] = None,
    additional_metrics: Optional[
        Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
    ] = None,
    verbose_level: Literal[None, 0, 1, 2] = 1,
    seed: Optional[int] = None,
    max_iter_no_improve: Optional[int] = None,
) -> Dict[str, object]:
    r"""Performs the training of the neural network ``model`` to fit the provided training data.

    Parameters
    ----------
    model : NeuralNetwork
        model to train.
    dataset : Union[RegressionDataset, Tuple[RegressionDataset, RegressionDataset]]
        dataset to use for training and validation. This argument is used with ``mask_dataset`` (to define the corresponding masked values) and ``val_frac`` (to define the proportion of entries to use in the validation set).
    learning_parameters : Union[LearningParameters, List[LearningParameters]]
        parameters of the stochastic gradient descent algorithm.
    mask_dataset : Union[ MaskDataset, Tuple[MaskDataset, MaskDataset], None, Tuple[None, None] ], optional
        _description_, by default None
    train_samples : Optional[Sequence], optional
        samples to use for training. When used, the arguments ``dataset``, ``mask_dataset`` and ``val_frac`` are disregarded. By default None.
    val_samples : Optional[Sequence], optional
        samples to use for validation, by default None.
    val_frac : Optional[float], optional
        proportion of elements of the ``dataset`` to use in the validation set. If specified, should be between 0 and 1. By default None.
    additional_metrics: Optional[Dict[str, Callable[[torch.Tensor, torch.Tensor], float]]], optional
        metrics to track in addition to the loss function, by default None
    verbose_level : Literal[None, 0, 1, 2], optional
        amount of information provided during training. 0 or None: no display. 1: display of the epoch bar. 2: display of network description and also epoch and batch bars. Default: 1.
    seed : Optional[int], optional
        random seed for reproducibility, by default None.
    max_iter_no_improve : Optional[int], optional
        early stopping parameter. The training stops when the loss on the validation set does not decrease for ``max_iter_no_improve`` steps. By default None.

    Returns
    -------
    Dict[str, object]
        This dictionary contains:
        - "train_loss": the time series of the average train loss per epoch.
        - "val_loss": the time series of the validation loss per epoch.
        - "train_metrics": the time series of additional metrics on train set.
        - "val_metrics": the time series of additional metrics on test set.
        - "train_set": train_set.
        - "val_set": val_set.
        - "lr": the time series of the learning rate per epoch.
        - "batch_size": the time series of the batch size per epoch.
        - "duration": total duration of training.

    Raises
    ------
    TypeError
        The ``dataset`` argument must be an instance of "RegressionDataset" or a tuple of two "RegressionDataset".
    TypeError
        The ``mask_dataset`` argument must be an instance of "MaskDataset" or a tuple of two "MaskDataset" or None.
    ValueError
        The ``mask_dataset`` argument must not be a tuple when the ``dataset`` argument is a "RegressionDataset".
    ValueError
        The ``dataset`` argument must not be a tuple when ``mask_dataset`` is a "MaskDataset".
    ValueError
        The train dataset and validation dataset must not share samples.
    ValueError
        The training dataset must not contain non finite values.
    ValueError
        The ``learning_parameters`` argument must be an instance of "LearningParameters" or a list of "LearningParameters".
    ValueError
        The ``learning_parameter.loss_function`` must not be an instance of "MaskedLossFunction" when ``mask_dataset=None``.
    """

    # Start counter
    tic = datetime.datetime.now()

    if not isinstance(model, NeuralNetwork):
        raise TypeError(
            f"model must be an instance of NeuralNetwork, not {type(model)}"
        )

    if isinstance(dataset, RegressionDataset):
        pass
    elif (
        isinstance(dataset, Sequence)
        and len(dataset) == 2
        and isinstance(dataset[0], RegressionDataset)
        and isinstance(dataset[1], RegressionDataset)
    ):
        pass
    else:
        raise TypeError(
            f"dataset must be an instance of RegressionDataset or a tuple of two RegressionDataset, not {type(dataset)}"
        )

    if isinstance(mask_dataset, MaskDataset) or mask_dataset is None:
        pass
    elif (
        isinstance(mask_dataset, Sequence)
        and len(mask_dataset) == 2
        and mask_dataset[0] is None
        and mask_dataset[1] is None
    ):
        mask_dataset = None
    elif (
        isinstance(mask_dataset, Sequence)
        and len(mask_dataset) == 2
        and isinstance(mask_dataset[0], MaskDataset)
        and isinstance(mask_dataset[1], MaskDataset)
    ):
        pass
    else:
        raise TypeError(
            f"mask_dataset must be an instance of MaskDataset or a tuple of two MaskDataset or None, not {type(mask_dataset)}"
        )

    if isinstance(dataset, RegressionDataset) and isinstance(mask_dataset, Sequence):
        raise ValueError(
            "mask_dataset must not be a tuple when dataset is a RegressionDataset"
        )
    if isinstance(dataset, Sequence) and isinstance(mask_dataset, MaskDataset):
        raise ValueError(
            "dataset must not be a tuple when mask_dataset is a MaskDataset"
        )

    if not isinstance(learning_parameters, LearningParameters):
        raise TypeError(
            f"learning_parameters must be an instance of LearningParameters, not {type(learning_parameters)}"
        )

    if additional_metrics is None:
        additional_metrics = {}

    if verbose_level is None:
        verbose_level = 0
    elif verbose_level in [0, 1, 2]:
        pass
    elif verbose_level in ["0", "1", "2"]:
        verbose_level = int(verbose_level)
    else:
        verbose_level = 1  # Default

    if max_iter_no_improve is not None:
        assert isinstance(max_iter_no_improve, int)
        assert max_iter_no_improve >= 1

    if seed is not None:
        random.seed(seed)

    if isinstance(dataset, RegressionDataset):
        if train_samples is not None and val_samples is not None:
            pass

        if train_samples is not None and val_samples is None:
            pass

        if train_samples is None and val_samples is None and val_frac is not None:
            n_val = round(val_frac * len(dataset))
            indices = list(range(len(dataset)))
            random.shuffle(indices)
            train_samples, val_samples = indices[n_val:], indices[:n_val]

        if train_samples is None and val_samples is None and val_frac is None:
            train_samples, val_samples = list(range(len(dataset))), None

        if train_samples is not None and val_samples is not None:
            intersect = set(train_samples) & set(val_samples)
            if len(intersect) > 0:
                raise ValueError(
                    "Train dataset and validation dataset must not share samples, here {intersect}"
                )

        train_set = RegressionSubset(dataset, train_samples)
        val_set = (
            RegressionSubset(dataset, val_samples) if val_samples is not None else None
        )

        if mask_dataset is not None:
            train_mask = MaskSubset(mask_dataset, train_samples)
            val_mask = (
                MaskSubset(mask_dataset, val_samples)
                if val_samples is not None
                else None
            )

    else:
        train_set = dataset[0]
        val_set = dataset[1]

        if mask_dataset is not None:
            train_mask = mask_dataset[0]
            val_mask = mask_dataset[1]

    if any(train_set.has_nonfinite()):
        raise ValueError("Non finite values in training dataset")

    # Training loop

    if verbose_level >= 2:
        count = model.count_parameters()
        size, unit = model.count_bytes()
        print("Training initiated")
        print(
            f"{model}: {count:,} learnable parameters ({size:.2f} {unit})", end="\n\n"
        )

    if isinstance(learning_parameters, LearningParameters):
        learning_parameters = [learning_parameters]
    elif not isinstance(learning_parameters, (list, tuple)):
        raise ValueError(
            "learning_parameters must be an instance of LearningParameters or a list of Learning Parameters"
        )
    elif any(not isinstance(p, LearningParameters) for p in learning_parameters):
        raise ValueError(
            "learning_parameters must be an instance of LearningParameters or a list of Learning Parameters"
        )

    train_loss = []
    val_loss = []
    train_metrics = dict.fromkeys(additional_metrics)
    for key in train_metrics:
        train_metrics[key] = []
    val_metrics = dict.fromkeys(additional_metrics)
    for key in val_metrics:
        val_metrics[key] = []
    lr = []
    bs = []

    for learning_parameter in learning_parameters:
        epochs = learning_parameter.epochs
        batch_size = learning_parameter.batch_size
        if batch_size is None:
            batch_size = len(train_set)
        if isinstance(batch_size, BatchScheduler):
            if batch_size.start is None:
                batch_size.start = len(train_set)  # By default, non-stochastic
            if batch_size.stop is None:
                batch_size.stop = len(train_set)  # By default, non-stochastic
        optimizer = learning_parameter.optimizer
        scheduler = learning_parameter.scheduler

        loss_fun = learning_parameter.loss_fun
        if mask_dataset is not None and not isinstance(loss_fun, MaskedLossFunction):
            loss_fun = MaskOverlay(loss_fun)
        if mask_dataset is None and isinstance(loss_fun, MaskedLossFunction):
            raise ValueError(
                "learning_parameter.loss_function must not be an instance of MaskedLossFunction when mask_dataset is None"
            )

        # Dataloaders

        if isinstance(batch_size, BatchScheduler):
            _batch_size = batch_size.get_batch_size()
        else:
            _batch_size = batch_size

        dataloader_train = DataLoader(
            train_set
            if mask_dataset is None
            else TensorDataset(train_set.x, train_set.y, train_mask.m),
            _batch_size,
            shuffle=True,
            drop_last=True,
            pin_memory=True,
        )
        dataloader_train_eval = DataLoader(
            train_set
            if mask_dataset is None
            else TensorDataset(train_set.x, train_set.y, train_mask.m),
            len(train_set),
            shuffle=False,
            drop_last=False,
            pin_memory=True,
        )
        dataloader_val_eval = DataLoader(
            val_set
            if mask_dataset is None
            else TensorDataset(val_set.x, val_set.y, val_mask.m),
            len(val_set),
            shuffle=False,
            drop_last=False,
            pin_memory=True,
        )

        n_batchs_train = len(dataloader_train)
        n_batchs_train_eval = 1
        n_batchs_val_eval = 1

        pbar_epoch = tqdm(range(epochs), disable=verbose_level < 1)
        pbar_epoch.set_description("Epoch")

        for epoch in pbar_epoch:
            # Dataloader
            if isinstance(batch_size, BatchScheduler):
                _batch_size = batch_size.get_batch_size()

                dataloader_train = DataLoader(
                    train_set
                    if mask_dataset is None
                    else TensorDataset(train_set.x, train_set.y, train_mask.m),
                    _batch_size,
                    shuffle=True,
                    drop_last=True,
                    pin_memory=True,
                )

            n_batchs_train = len(dataloader_train)

            lr.append(optimizer.param_groups[0]["lr"])
            bs.append(_batch_size)

            # Training
            model.train()

            pbar_batch = tqdm(
                enumerate(dataloader_train),
                leave=False,
                total=n_batchs_train,
                disable=verbose_level < 2,
            )
            pbar_batch.set_description("Batch (training)")
            for _, batch in pbar_batch:
                optimizer.zero_grad(set_to_none=True)

                loss, _ = _batch_processing(
                    model,
                    batch,
                    loss_fun,
                    mask_dataset is None,
                    metrics=additional_metrics,
                )

                loss.backward()
                optimizer.step()

                pbar_batch.set_postfix({"loss": loss.item()})

            model.eval()

            # Evaluation on train set

            sizes = []
            memory_loss = []
            memory_metrics = dict.fromkeys(train_metrics.keys())
            for key in memory_metrics:
                memory_metrics[key] = []

            pbar_batch = tqdm(
                enumerate(dataloader_train_eval),
                leave=False,
                total=n_batchs_train_eval,
                disable=verbose_level < 1,
            )
            pbar_batch.set_description("Batch (model eval)")
            for _, batch in pbar_batch:
                sizes.append(batch[0].size(0))

                with torch.no_grad():
                    loss, mets = _batch_processing(
                        model,
                        batch,
                        loss_fun,
                        mask_dataset is None,
                        metrics=additional_metrics,
                    )

                memory_loss.append(loss.item())
                for key in memory_metrics:
                    memory_metrics[key].append(mets[key].item())

            n_tot = sum(sizes)
            train_loss.append(sum([s / n_tot * l for l, s in zip(memory_loss, sizes)]))
            for key in train_metrics:
                train_metrics[key].append(
                    sum([s / n_tot * val for val, s in zip(memory_metrics[key], sizes)])
                )

            # Evaluation on validation set

            sizes = []
            memory_loss = []
            memory_metrics = dict.fromkeys(val_metrics.keys())
            for key in memory_metrics:
                memory_metrics[key] = []

            pbar_batch = tqdm(
                enumerate(dataloader_val_eval),
                leave=False,
                total=n_batchs_val_eval,
                disable=verbose_level < 2,
            )
            pbar_batch.set_description("Batch (validation)")

            for _, batch in pbar_batch:
                sizes.append(batch[0].size(0))

                with torch.no_grad():
                    loss, mets = _batch_processing(
                        model,
                        batch,
                        loss_fun,
                        mask_dataset is None,
                        metrics=additional_metrics,
                    )

                memory_loss.append(loss.item())
                for key in memory_metrics:
                    memory_metrics[key].append(mets[key].item())

            n_tot = sum(sizes)
            val_loss.append(sum([s / n_tot * l for l, s in zip(memory_loss, sizes)]))
            for key in val_metrics:
                val_metrics[key].append(
                    sum([s / n_tot * val for val, s in zip(memory_metrics[key], sizes)])
                )

            # End of epoch
            if isinstance(scheduler, ReduceLROnPlateau):
                scheduler.step(train_loss[-1])
            else:
                scheduler.step()
            if isinstance(batch_size, BatchScheduler):
                batch_size.step()

            if len(additional_metrics) > 0:
                key = list(additional_metrics.keys())[
                    0
                ]  # Display only the first metric
                add = {
                    f"train {key}": train_metrics[key][-1],
                    f"val  {key}": val_metrics[key][-1],
                }
            else:
                add = {}
            pbar_epoch.set_postfix(
                {
                    "train loss": train_loss[-1],
                    "val loss": val_loss[-1],
                }.update(add)
            )

            # Early stopping
            if (
                max_iter_no_improve is not None
                and epoch > max_iter_no_improve
                and np.min(train_loss[-max_iter_no_improve:]) > np.min(train_loss)
            ):
                break

    model.eval()

    toc = datetime.datetime.now()

    print()

    return {
        "train_loss": train_loss,
        "val_loss": val_loss,
        "train_metrics": train_metrics,
        "val_metrics": val_metrics,
        "train_set": train_set,
        "val_set": val_set,
        "lr": lr,
        "batch_size": bs,
        "duration": toc - tic,
    }


def _batch_processing(
    model: NeuralNetwork,
    batch: Optional[torch.Tensor],
    loss_fun: Callable,
    masked: bool,
    metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]],
) -> Tuple[torch.Tensor, Optional[Dict[str, torch.Tensor]]]:
    if masked:
        x, y = batch
        m = None
    else:
        x, y, m = batch

    # Get the data to GPU (if available)
    x = x.to(model.device, non_blocking=True)
    y = y.to(model.device, non_blocking=True)

    if m is not None:
        m = m.to(model.device, non_blocking=True)

    y_hat = model.forward(x)

    if m is None:
        loss = loss_fun(y_hat, y)
    else:
        loss = loss_fun(y_hat, y, m)

    mets = dict.fromkeys(metrics.keys())
    for key in mets:
        if m is None:
            dist = metrics[key](y_hat.detach(), y)
            dist = dist.mean()
        else:
            w = 1.0 / m.sum(dim=0).clip(min=1)
            dist = metrics[key](y_hat.detach(), y)
            dist = (w * dist).sum(dim=0).mean()
        mets[key] = dist

    return loss, mets



import numpy as np

__all__ = ["id", "log10", "pow10", "asinh", "sinh"]


def id(t: np.ndarray):
    return t


def log10(t: np.ndarray):
    return np.log10(t)


def pow10(t: np.ndarray):
    return 10**t


def asinh(t: np.ndarray, a: float = 1.0):
    return a * np.arcsinh(t / a)


def sinh(t: np.ndarray, a: float = 1.0):
    return a * sinh(t / a)



from itertools import accumulate, chain
from typing import Optional, Sequence, Union
from warnings import warn

import torch
from torch import Tensor, nn

from ..operators import Operator
from .neural_network import NeuralNetwork

__all__ = ["MergingNetwork"]


class MergingNetwork(NeuralNetwork):
    r"""Utility class to run a set of neural networks in parallel to predict distinct sets of outputs."""

    def __init__(
        self,
        subnetworks: Sequence[NeuralNetwork],
        inputs_names: Optional[Sequence[str]] = None,
        outputs_names: Optional[Sequence[str]] = None,
        inputs_transformer: Optional[Operator] = None,
        outputs_transformer: Optional[Operator] = None,
        device: Optional[str] = None,
    ):
        """

        Parameters
        ----------
        subnetworks : Sequence[NeuralNetwork]
            set of neural networks to be run in parallel to predict distinct sets of outputs.
        inputs_names : Optional[Sequence[str]], optional
            List of inputs names. None if the names have not been specified. By default None.
        outputs_names : Optional[Sequence[str]], optional
            List of outputs names. None if the names have not been specified. By default None.
        inputs_transformer : Optional[Operator], optional
            Transformation applied to the inputs before processing, by default None.
        outputs_transformer : Optional[Operator], optional
            Transformation applied to the outputs after processing, by default None.
        device : Optional[str], optional
            Device used ("cpu" or "cuda"), by default None (corresponds to "cpu").

        Raises
        ------
        TypeError
            The ``subnetworks`` argument must be a sequence of NeuralNetwork instances.
        ValueError
            All the elements of ``subnetworks`` must have the same number of inputs.
        ValueError
            Incompatible ``inputs_names`` among ``subnetworks``.
        ValueError
            No element of ``subnetworks`` can be None when outputs_names is not None.
        ValueError
            Some elements of ``subnetworks`` have the same outputs.
        ValueError
            Some elements of ``outputs_names`` cannot be found in the outputs of any element of ``subnetworks``.
        """
        # Subnetworks
        if any([not isinstance(net, NeuralNetwork) for net in subnetworks]):
            raise TypeError("subnetworks must be a sequence of NeuralNetwork")

        # Inputs
        n_inputs = subnetworks[0].input_features
        if any([net.input_features != n_inputs for net in subnetworks]):
            raise ValueError(
                "All element of subnetworks must have the same number of inputs"
            )

        _inputs_names = [
            net.inputs_names for net in subnetworks if net.inputs_names is not None
        ]
        if any(name != _inputs_names[0] for name in _inputs_names):
            raise ValueError("Incompatible inputs_names among subnetworks")
        inputs_names = _inputs_names[0] if len(_inputs_names) > 0 else None

        # Outputs
        n_outputs = sum([net.output_features for net in subnetworks])
        if outputs_names is not None and any(
            [net.outputs_names is None for net in subnetworks]
        ):
            raise ValueError(
                "No element of subnetwork can be None when outputs_names is not None"
            )

        if outputs_names is None:
            if all([net.outputs_names is not None for net in subnetworks]):
                outputs_names = list(chain(*[net.outputs_names for net in subnetworks]))
            self.indices = list(
                range(sum([net.output_features for net in subnetworks]))
            )
        else:
            _outputs_names = list(chain(*[net.outputs_names for net in subnetworks]))
            if len(set(_outputs_names)) != len(_outputs_names):
                raise ValueError("Some subnetworks have the same outputs")
            if not (set(_outputs_names) >= set(outputs_names)):
                raise ValueError(
                    "Some elements of outputs_names cannot be found in any subnetworks"
                )
            if set(_outputs_names) != set(outputs_names):
                warn("Some subnetworks outputs are not retrieved")
            if len(_outputs_names) != len(outputs_names):
                warn("There are duplicates in outputs_names")
            self.indices = [_outputs_names.index(name) for name in outputs_names]

        super().__init__(
            n_inputs,
            n_outputs,
            inputs_names=inputs_names,
            outputs_names=outputs_names,
            inputs_transformer=inputs_transformer,
            outputs_transformer=outputs_transformer,
            device=device,
        )

        self.subnetworks = nn.ModuleList(subnetworks)

    def forward(self, x: Tensor) -> Tensor:
        res = []

        for net in self.subnetworks:
            res.append(net.forward(x))

        return torch.concat(res, dim=-1)[..., self.indices]

    def restrict_to_output_subset(
        self, output_subset: Optional[Union[Sequence[str], Sequence[int]]]
    ) -> None:
        super().restrict_to_output_subset(output_subset)

        if isinstance(output_subset[0], int):
            net_start_indices = list(
                accumulate([0] + [net.output_features for net in self.subnetworks][:-1])
            )
            net_end_indices = list(
                accumulate([net.output_features for net in self.subnetworks])
            )

        for net in self.subnetworks:
            if output_subset is None:
                net_output_subset = None
            elif isinstance(output_subset[0], str):
                net_output_subset = [
                    name for name in output_subset if name in net.outputs_names
                ]
            else:
                start, end = next(net_start_indices), next(net_end_indices)
                real_output_subset = [
                    self.indices[idx]
                    for idx in output_subset
                    if idx >= start and idx < end
                ]
                net_output_subset = [idx - start for idx in real_output_subset]
            net.restrict_to_output_subset(net_output_subset)



from math import ceil
from typing import Optional, Sequence, Union

from torch import Tensor, concat, nn

from ..layers import RestrictableLinear
from ..operators import Operator
from .neural_network import NeuralNetwork

__all__ = ["DenselyConnected"]


class DenselyConnected(NeuralNetwork):
    r"""Densely connected neural network. In such a network, the input of an hidden layer is the concatenation of the input and output of the previous layer. This `skip` operation permits to reduce the number of parameters to learn, to reuse intermediate computation results and to avoid gradient vanishing effects."""

    def __init__(
        self,
        input_features: int,
        output_features: int,
        n_layers: int,
        growing_factor: float,
        activation: nn.Module,
        batch_norm: bool = False,
        inputs_names: Optional[Sequence[str]] = None,
        outputs_names: Optional[Sequence[str]] = None,
        inputs_transformer: Optional[Operator] = None,
        outputs_transformer: Optional[Operator] = None,
        device: Optional[str] = None,
        last_restrictable: bool = True,
    ):
        """

        Parameters
        ----------
        input_features : int
            dimension of input vector.
        output_features : int
            dimension of output vector.
        n_layers : int
            number of layers in the network.
        growing_factor : float
            growing factor considered in the full network. The growing factor corresponds to the ratio of the output and input dimensions for one layer. For instance, ``growing_factor=1.0`` implies that the input of a hidden layer is twice that of the previous layer.
        activation : nn.Module
            activation function.
        batch_norm : bool, optional
            whether to use batch normalization during training, by default ``False``.
        inputs_names : Optional[Sequence[str]], optional
            List of inputs names. None if the names have not been specified. By default None.
        outputs_names : Optional[Sequence[str]], optional
            List of outputs names. None if the names have not been specified. By default None.
        inputs_transformer : Optional[Operator], optional
            Transformation applied to the inputs before processing, by default None.
        outputs_transformer : Optional[Operator], optional
            Transformation applied to the outputs after processing, by default None.
        device : Optional[str], optional
            Device used ("cpu" or "cuda"), by default None (corresponds to "cpu").
        last_restrictable : bool, optional
            whether the last layer is to be a RestrictableLinear layer, by default ``True``.
        """
        super().__init__(
            input_features,
            output_features,
            inputs_names=inputs_names,
            outputs_names=outputs_names,
            inputs_transformer=inputs_transformer,
            outputs_transformer=outputs_transformer,
            device=device,
        )

        self.n_layers = n_layers
        self.growing_factor = growing_factor

        self.activation = activation
        self.batch_norm = batch_norm

        self.last_restrictable = last_restrictable

        self.layers = nn.ModuleList()
        n_inputs = input_features
        self.layers_sizes = [input_features]
        for k in range(n_layers - 1):
            n_outputs = ceil(growing_factor * n_inputs)

            if batch_norm and k < n_layers - 2:
                self.layers.append(
                    nn.Sequential(
                        nn.Linear(n_inputs, n_outputs, device=self.device),
                        nn.BatchNorm1d(n_outputs, device=self.device),
                    )
                )
            else:
                self.layers.append(nn.Linear(n_inputs, n_outputs, device=self.device))

            n_inputs += n_outputs
            self.layers_sizes.append(n_inputs)

        if last_restrictable:
            self.output_layer = RestrictableLinear(
                n_inputs,
                output_features,
                outputs_names=self.outputs_names,
                device=self.device,
            )
        else:
            self.output_layer = nn.Linear(
                n_inputs,
                output_features,
                device=self.device,
            )
        self.layers_sizes.append(output_features)

    def forward(self, x: Tensor) -> Tensor:
        xk = x.clone()

        is1d = xk.ndim == 1
        if is1d:
            xk = xk.unsqueeze(0)

        for layer in self.layers:
            yk = layer(xk)
            yk = self.activation(yk)
            xk = concat((xk, yk), axis=-1)

        y_hat = self.output_layer(xk)

        if not self.last_restrictable:
            y_hat = y_hat[..., self.current_output_subset_indices]
        if is1d:
            y_hat = y_hat.squeeze(0)

        return y_hat

    def restrict_to_output_subset(
        self, output_subset: Optional[Union[Sequence[str], Sequence[int]]]
    ) -> None:
        super().restrict_to_output_subset(output_subset)
        if self.last_restrictable:
            self.output_layer.restrict_to_output_subset(
                self.current_output_subset_indices
            )



import os
import pickle
from typing import Dict, List, Literal, Optional, Sequence, Tuple, Union, overload

import numpy as np
import pandas as pd
from torch import Tensor, from_numpy
from torch.utils.data import Dataset

__all__ = ["MaskDataset", "MaskSubset"]


class MaskDataset(Dataset):
    """Dataset dedicated to ignore some specified values during learning."""

    def __init__(
        self,
        m: np.ndarray,
        features_names: Optional[List[str]] = None,
    ):
        r"""

        Parameters
        ----------
        m : np.ndarray
            Array containing the output features of the regression model.
            y must be of shape :math:`N \times F` where :math:`N` is the number of entries and :math:`F` the number of features.
        features_names : Optional[List[str]], optional
            list of feature names, by default None.

        Raises
        ------
        ValueError
            ``m`` and ``features_names`` must have the same number of features :math:`F`.
        """
        super().__init__()

        self._m: Tensor = from_numpy(m).float()

        if features_names is not None and len(features_names) != m.shape[1]:
            raise ValueError(
                "m and features_names must have the same number of features"
            )

        self._features_names: Optional[List[str]] = features_names

    def __len__(self) -> int:
        r"""
        Returns the number of entries :math:`N` in the dataset.

        Returns
        -------
        int
            Number of entries :math:`N`.
        """
        return self._m.size(0)

    def __getitem__(self, idx) -> Tuple[Tensor, Tensor]:
        r"""
        Returns the index entries idx.

        Parameters
        ----------
        idx : Any
            Indices of entries to return.

        Returns
        -------
        tuple of torch.Tensor
            Input and output entries.
        """
        return self._m[idx]

    @property
    def m(self) -> Tensor:
        r"""
        Mask Tensor.
        """
        return self._m

    @property
    def n_features(self) -> int:
        r"""
        Number of output features.
        """
        return self.m.size(1)

    @property
    def features_names(self) -> Optional[List[str]]:
        r"""
        Features names.
        """
        return self._features_names

    def features_size(self) -> int:
        """
        Returns the number of floating point values in ``m``.
        """
        return self.m.numel()

    @overload
    def getall(self, numpy: Literal[True]) -> np.ndarray:
        ...

    @overload
    def getall(self, numpy: Literal[False]) -> Tensor:
        ...

    def getall(self, numpy: bool = False) -> Union[np.ndarray, Tensor]:
        r"""
        Returns all the dataset in numpy.ndarray or torch.Tensor depending on the value of the ``numpy`` parameter.

        Parameters
        ----------
        numpy : bool, optional
            If ``numpy==True``, the returned object will be numpy arrays.
            Else, they will be torch tensors.

        Returns
        -------
        torch.Tensor or numpy.ndarray
            Mask.
        """
        m = self[list(range(len(self)))]
        if numpy:
            return m.numpy()
        return m

    @staticmethod
    def from_pandas(df_m: pd.DataFrame) -> "MaskDataset":
        r"""Converts a pandas DataFrame to a MaskDataset object.

        Parameters
        ----------
        df_m : pd.DataFrame
            DataFrame of the masked outputs. This DataFrame should contain :math:`N` rows, i.e., number of entries, and :math:`F` columns, i.e., features.

        Returns
        -------
        MaskDataset
            associated MaskDataset object. The ``m`` attribute is set to values in the DataFrame, and the ``feature_names`` attribute to the column names.
        """
        return MaskDataset(
            df_m.values,
            df_m.columns.to_list(),
        )

    def to_pandas(self) -> pd.DataFrame:
        r"""Converts the mask dataset to a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
            DataFrame of the mask on the output y.
        """
        return pd.DataFrame(self.m, columns=self._features_names)

    def join(self: "MaskDataset", other: "MaskDataset") -> "MaskDataset":
        r"""
        Returns the union of two datasets. Data are copied.

        Parameters
        ----------
        other : MaskDataset
            Other dataset to join with.

        Returns
        -------
        MaskDataset
            New dataset constructed as the union of the two datasets.
        """
        x1, y1 = self.getall(numpy=True)
        x2, y2 = other.getall(numpy=True)
        x = np.concatenate((x1, x2), axis=0)
        y = np.concatenate((y1, y2), axis=0)
        return MaskDataset(x, y)

    def substract(self: "MaskDataset", other: "MaskSubset") -> "MaskDataset":
        r"""
        Returns the subtraction of two datasets. Data are copied.

        Parameters
        ----------
        other : MaskDataset
            Subset of ``self``.

        Returns
        -------
        MaskDataset
            New subset of ``self`` containing all values that were not in `other`.
        """
        if not other.issubsetof(self):
            raise ValueError("set2 is not a subset of set1 so it cannot be subtracted.")
        new_indices = [i for i in range(len(self)) if i not in other.indices]
        # Algo can be improved.
        return MaskDataset(self, new_indices)

    def stats(self) -> Dict[str, np.ndarray]:
        r"""Computes the proportion of masked entries for each output column.

        Returns
        -------
        Dict[str, np.ndarray]
            dictionary of masked entry proportion for each output feature.
        """
        return {
            "frac": self.m.mean(axis=0).numpy(),
        }

    def save(self, filename: str, path: Optional[str] = None) -> None:
        r"""saves the dataset to a pickle file.

        Parameters
        ----------
        filename : str
            name of the file to be created.
        path : Optional[str], optional
            path to the file to be created, by default None.
        """
        if path is not None:
            filename = os.path.join(path, filename)
        filename = os.path.splitext(filename)[0]
        with open(f"{filename}.pkl", "wb") as file:
            pickle.dump(self, file)

    @staticmethod
    def load(filename: str, path: Optional[str] = None) -> "MaskDataset":
        r"""loads a mask dataset from a pickle file.

        Parameters
        ----------
        filename : str
            name of the file to be read.
        path : Optional[str], optional
            path to the file to be read, by default None.

        Returns
        -------
        MaskDataset
            loaded mask dataset.
        """
        if path is not None:
            filename = os.path.join(path, filename)
        filename = os.path.splitext(filename)[0]
        with open(f"{filename}.pkl", "rb") as file:
            dataset = pickle.load(file)
        return dataset


class MaskSubset(MaskDataset):
    r"""Subset of RegressionDataset."""

    def __init__(self, dataset: MaskDataset, indices: Sequence[int]):
        r"""

        Parameters
        ----------
        dataset : MaskDataset
            Dataset from which entries are extracted.
        indices : Sequence of int
            Indices of entries to extract.
        """
        self._dataset: MaskDataset = dataset
        self._indices: Sequence[int] = indices

    def __len__(self) -> int:
        r"""
        Returns the number of entries in the dataset.

        Returns
        -------
        int
            Number of entries.
        """
        return len(self._indices)

    def __getitem__(self, idx) -> Tensor:
        r"""
        Returns the index entries idx.

        Parameters
        ----------
        idx : Any
            Indices of entries to return.

        Returns
        -------
        tuple of torch.Tensor
            Input and output entries.
        """
        return self._dataset[self._indices[idx]]

    @property
    def m(self) -> Tensor:
        r"""
        Mask Tensor.
        """
        return self._dataset._m[self._indices]

    def issubsetof(self, dataset: MaskDataset) -> bool:
        r"""
        Returns ``True`` if ``self`` is a subset of ``dataset``.

        Parameters
        ----------
        dataset : MaskDataset
            Dataset of which we want to know if ``self`` is a subset.

        Returns
        -------
        bool
            ``True`` if ``self`` is a subset of ``dataset`` else ``False``.
        """
        # TODO maybe a better solution is suitable
        return dataset == self._dataset



import math
from abc import ABC, abstractmethod

import torch
from torch import nn

LOG10 = math.log(10.0)

__all__ = [
    "MaskedLossFunction",
    "MaskOverlay",
    "MaskedMSELoss",
    "CauchyLoss",
    "SmoothL1Loss",
]

## Masked loss functions


class MaskedLossFunction(nn.Module, ABC):
    r"""Implements a masked loss function which has a signature ``loss_fun(y_hat, y, mask)``."""

    def __init__(self):
        r""" """
        super().__init__()

    @abstractmethod
    def forward(
        y_hat: torch.Tensor, y: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        r"""Evaluates the loss between a prediction ``y_hat`` and a reference ``y`` with some masked values indicated in ``mask``.

        Parameters
        ----------
        y_hat : torch.Tensor
            network prediction.
        y : torch.Tensor
            true values. Must have the same shape as ``y_hat``.
        mask : torch.Tensor
            binary mask with values to disregard in the loss. Must have the same shape as ``y_hat``.

        Returns
        -------
        torch.Tensor
            evaluated loss, should be a float.
        """
        pass


class MaskOverlay(MaskedLossFunction):
    r"""Permits to use a MaskedLossFunction as a standard loss function."""

    def __init__(self, loss):
        r"""

        Parameters
        ----------
        loss : Callable
            loss function
        """
        super().__init__()
        self.loss = loss

    def forward(self, y_hat: torch.Tensor, y: torch.Tensor, _) -> torch.Tensor:
        """Evaluate the loss function.

        Parameters
        ----------
        y_hat : torch.Tensor
            network prediction.
        y : torch.Tensor
            true values. Must have the same shape as ``y_hat``.

        Returns
        -------
        torch.Tensor
            evaluated loss, should be a float.
        """
        return self.loss(y_hat, y)


class MaskedMSELoss(MaskedLossFunction):
    r"""Implements the masked MSE loss function, i.e., for a binary mask :math:`m`, a prediction :math:`\widehat{y}` and a true value :math:`y`,


    .. math::

       \mathrm{MaskedMSE}(\widehat{y}, y, m) = \frac{1}{\sum_{i} m_{i}} \sum_{i} m_{i} \left( \widehat{y}_{i} - y_{i} \right)^2

    """

    def __init__(self):
        r""" """
        super().__init__()

    def forward(
        self, y_hat: torch.Tensor, y: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        r"""Evaluates the masked MSE loss between a prediction ``y_hat`` and a reference ``y``, with a binary mask ``m`` -- where :math:`m_{i}=0` corresponds to a masked value.

        Parameters
        ----------
        y_hat : torch.Tensor
            network prediction.
        y : torch.Tensor
            true values. Must have the same shape as ``y_hat``.
        mask : torch.Tensor
            binary mask. Must have the same shape as ``y_hat``.

        Returns
        -------
        torch.Tensor
            evaluated loss, should be a float.
        """
        w = mask / mask.sum(dim=0).clip(min=1)
        return (w * (y_hat - y).square()).sum(dim=0).mean()


# class MaskedWeightedMSELoss(MaskedLossFunction):
#     def __init__(self, w: List):
#         super().__init__()
#         self.w: torch.Tensor = torch.tensor(w).flatten()

#     def forward(
#         self, y_hat: torch.Tensor, y: torch.Tensor, mask: torch.Tensor
#     ) -> torch.Tensor:
#         w = mask / mask.sum(dim=0).clip(min=1)
#         return (self.w * w * (y_hat - y).square()).sum(dim=0).mean()


# class MaskedMAELoss(MaskedLossFunction):
#     def __init__(self):
#         super().__init__()

#     def forward(self, y_hat: torch.Tensor, y: torch.Tensor, mask: torch.Tensor):
#         w = mask / mask.sum(dim=0).clip(min=1)
#         return (w * (y_hat - y).abs()).sum(dim=0).mean()


# class MaskedPowerLoss(MaskedLossFunction):
#     degree: int

#     def __init__(self, degree: int):
#         super().__init__()
#         self.degree = degree

#     def forward(self, y_hat: torch.Tensor, y: torch.Tensor, mask: torch.Tensor):
#         w = mask / mask.sum(dim=0).clip(min=1)
#         return (w * (y_hat - y).abs() ** self.degree).sum(dim=0).mean()


# class MaskedSeriesLoss(MaskedLossFunction):
#     def __init__(self, order: int):
#         super().__init__()
#         self.order = order

#     def forward(
#         self, y_hat: torch.Tensor, y: torch.Tensor, mask: torch.Tensor
#     ) -> torch.Tensor:
#         w = mask / mask.sum(dim=0).clip(min=1)
#         diff = (y_hat - y).abs()
#         err = 0.0 * diff
#         logk = 1.0
#         diffk = 0.0 * diff + 1.0
#         denomk = 1
#         for k in range(1, self.order + 1):
#             logk = logk * LOG10
#             diffk = diffk * diff
#             denomk = denomk * k
#             err = err + logk * diffk / denomk
#         return (w * err).sum(dim=0).mean()


# class MaskedRelErrorLoss(MaskedLossFunction):
#     def __init__(self):
#         super().__init__()

#     def forward(self, y_hat: torch.Tensor, y: torch.Tensor, mask: torch.Tensor):
#         w = mask / mask.sum(dim=0).clip(min=1)
#         relerr = (LOG10 * (y_hat - y).abs()).exp() - 1
#         return (w * relerr).sum(dim=0).mean()


# Custom classic loss functions


class CauchyLoss(nn.Module):
    r"""Implements the Cauchy loss function, i.e.,


    .. math::

       \mathrm{CL}(\widehat{y}, y) = \sum_{i} \log \left( 1 + \left( \widehat{y}_{i} - y_{i} \right)^2 \right)

    """

    def __init__(self):
        super().__init__()

    def forward(self, y_hat: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        r"""Evaluates the Cauchy loss between a prediction ``y_hat`` and a reference ``y``.

        Parameters
        ----------
        y_hat : torch.Tensor
            network prediction.
        y : torch.Tensor
            true values. Must have the same shape as ``y_hat``.

        Returns
        -------
        torch.Tensor
            evaluated loss, should be a float.
        """
        return torch.log(1 + (y_hat - y).square()).mean()


class SmoothL1Loss(nn.Module):
    r"""Implements the smooth L1 loss function, i.e.,


    .. math::

       \mathrm{SmoothL1}(\widehat{y}, y) = \sum_{i} \begin{cases} \frac{1}{2\beta}\left( \widehat{y}_{i} - y_{i} \right)^2 \; \text{ if } \vert \widehat{y}_{i} - y_{i} \vert \leq \beta \\ \vert \widehat{y}_{i} - y_{i} \vert - 0.5 \beta \; \text{ otherwise} \end{cases}

    """
    # beta: float

    def __init__(self, beta: float):
        r"""

        Parameters
        ----------
        beta : float
            :math:`\beta` parameter of the loss function.
        """
        super().__init__()
        self.beta = beta

    def forward(self, y_hat: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        r"""Evaluates the smooth L1 loss between a prediction ``y_hat`` and a reference ``y``.

        Parameters
        ----------
        y_hat : torch.Tensor
            network prediction.
        y : torch.Tensor
            true values. Must have the same shape as ``y_hat``.

        Returns
        -------
        torch.Tensor
            evaluated loss, should be a float.
        """
        abs_diffs = torch.abs(y_hat - y)
        return torch.mean(
            torch.where(
                abs_diffs < self.beta,
                0.5 * torch.square(abs_diffs) / self.beta,
                abs_diffs - 0.5 * self.beta,
            )
        )



from typing import Optional, Sequence, Union

from numpy import ndarray
from torch import Tensor

from ..layers import PolynomialExpansion
from ..operators import Operator
from .neural_network import NeuralNetwork

__all__ = ["PolynomialNetwork"]


class PolynomialNetwork(NeuralNetwork):
    r"""Neural network with a polynomial expansion as a first layer."""

    def __init__(
        self,
        input_features: int,
        order: int,
        subnetwork: NeuralNetwork,
        inputs_names: Optional[Sequence[str]] = None,
        outputs_names: Optional[Sequence[str]] = None,
        inputs_transformer: Optional[Operator] = None,
        outputs_transformer: Optional[Operator] = None,
        device: Optional[str] = None,
    ):
        r"""

        Parameters
        ----------
        input_features : int
            Dimension of input vector.
        order : int
            order of the polynomial expansion.
        subnetwork : NeuralNetwork
            network to be placed after the polynomial expansion.
        inputs_names : Optional[Sequence[str]], optional
            List of inputs names. None if the names have not been specified. By default None.
        outputs_names : Optional[Sequence[str]], optional
            List of outputs names. None if the names have not been specified. By default None.
        inputs_transformer : Optional[Operator], optional
            Transformation applied to the inputs before processing, by default None.
        outputs_transformer : Optional[Operator], optional
            Transformation applied to the outputs after processing, by default None.
        device : Optional[str], optional
            Device used ("cpu" or "cuda"), by default None (corresponds to "cpu").

        Raises
        ------
        ValueError
            The number of polynomial features does not match the input layer of the subnetwork.
        """
        super().__init__(
            input_features,
            subnetwork.output_features,
            inputs_names=inputs_names,
            outputs_names=outputs_names,
            inputs_transformer=inputs_transformer,
            outputs_transformer=outputs_transformer,
            device=device,
        )

        self.order = order
        self.n_poly_features = PolynomialExpansion.expanded_features(
            order, input_features
        )

        if self.n_poly_features != subnetwork.input_features:
            raise ValueError(
                f"The number of polynomial features ({self.n_poly_features}) does not match the input layer of the subnetwork ({subnetwork.input_features[0]})."
            )

        self.poly = PolynomialExpansion(input_features, order, self.device)
        self.subnetwork = subnetwork

    def forward(self, x: Tensor) -> Tensor:
        y_hat = self.poly.forward(x)
        y_hat = self.subnetwork.forward(y_hat)
        return y_hat

    def update_standardization(self, x: Union[Tensor, ndarray]) -> None:
        """Applies the ``update_standardization`` method of the PolynomialExpansion first layer, i.e., updates the standardization parameters for the outputs of the polynomial expansion.

        Parameters
        ----------
        x : Union[Tensor, ndarray]
            input tensor.
        """
        self.poly.update_standardization(x)

    def restrict_to_output_subset(
        self, output_subset: Optional[Union[Sequence[str], Sequence[int]]]
    ) -> None:
        super().restrict_to_output_subset(output_subset)
        self.subnetwork.restrict_to_output_subset(output_subset)



from typing import List, Optional, Sequence, Union

# issue with import for Python < 3.9
try:
    from itertools import pairwise
except:
    from more_itertools import pairwise

from torch import Tensor, nn

from ..layers import AdditionalModule
from ..operators import Operator
from .neural_network import NeuralNetwork

__all__ = ["EmbeddingNetwork"]


class EmbeddingNetwork(NeuralNetwork):
    r"""Embedding neural network."""

    def __init__(
        self,
        subnetwork: NeuralNetwork,
        preprocessing: Union[None, AdditionalModule, List[AdditionalModule]] = None,
        postprocessing: Union[None, AdditionalModule, List[AdditionalModule]] = None,
        inputs_names: Optional[Sequence[str]] = None,
        outputs_names: Optional[Sequence[str]] = None,
        inputs_transformer: Optional[Operator] = None,
        outputs_transformer: Optional[Operator] = None,
        device: Optional[str] = None,
    ):
        """

        Parameters
        ----------
        subnetwork : NeuralNetwork
            Base network.
        preprocessing : Union[None, AdditionalModule, List[AdditionalModule]], optional
            PyTorch operation to apply before ``subnetwork``, by default None.
        postprocessing : Union[None, AdditionalModule, List[AdditionalModule]], optional
            PyTorch operation to apply after ``subnetwork``, by default None.
        inputs_names : Optional[Sequence[str]], optional
            List of inputs names. None if the names have not been specified. By default None.
        outputs_names : Optional[Sequence[str]], optional
            List of outputs names. None if the names have not been specified. By default None.
        inputs_transformer : Optional[Operator], optional
            Transformation applied to the inputs before processing, by default None.
        outputs_transformer : Optional[Operator], optional
            Transformation applied to the outputs after processing, by default None.
        device : Optional[str], optional
            Device used ("cpu" or "cuda"), by default None (corresponds to "cpu").

        Raises
        ------
        TypeError
            All elements of preprocessing must be instances of AdditionalModule.
        TypeError
            All elements of postprocessing must be instances of AdditionalModule.
        """
        if preprocessing is None:
            preprocessing = []
        elif not isinstance(preprocessing, List):
            preprocessing = [preprocessing]
        if any([not isinstance(m, AdditionalModule) for m in preprocessing]):
            raise TypeError(
                "All elements of preprocessing must be instances of AdditionalModule"
            )

        n_inputs = subnetwork.input_features
        for m2, m1 in pairwise(([subnetwork] + preprocessing[::-1])):
            if (
                m2.input_features is not None
                and m1.output_features != m2.input_features
            ):
                raise ValueError(
                    f"{type(m1).__name__}.output_features ({m1.output_features}) doesn't match {type(m2).__name__}.input_features ({m1.input_features}"
                )
            if m1.input_features is not None:
                n_inputs = m1.input_features

        if postprocessing is None:
            postprocessing = []
        elif not isinstance(postprocessing, List):
            postprocessing = [postprocessing]
        if any([not isinstance(m, AdditionalModule) for m in postprocessing]):
            raise TypeError(
                "All elements of postprocessing must be instances of AdditionalModule"
            )

        n_outputs = subnetwork.output_features
        for m1, m2 in pairwise([subnetwork] + postprocessing):
            if (
                m2.output_features is not None
                and m1.output_features != m2.input_features
            ):
                raise ValueError(
                    f"{type(m1).__name__}.output_features ({m1.output_features}) doesn't match {type(m2).__name__}.input_features ({m2.input_features})"
                )
            if m2.output_features is not None:
                n_outputs = m2.output_features

        super().__init__(
            n_inputs,
            n_outputs,
            inputs_names=inputs_names,
            outputs_names=outputs_names,
            inputs_transformer=inputs_transformer,
            outputs_transformer=outputs_transformer,
            device=device,
        )

        self.subnetwork = subnetwork
        self.preprocessing = nn.ModuleList(preprocessing)
        self.postprocessing = nn.ModuleList(postprocessing)

    def forward(self, x: Tensor) -> Tensor:
        y = x.clone()
        for op in self.preprocessing:
            y = op(y)
        y = self.subnetwork.forward(y)
        for op in self.postprocessing:
            y = op(y)
        return y



from typing import Optional, Sequence

import torch
from torch import nn

__all__ = ["RestrictableLinear"]


class RestrictableLinear(nn.Linear):
    """
    Restrictable linear layer.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        device: Optional[str] = None,
        dtype=None,
        outputs_names: Optional[Sequence[str]] = None,
    ):
        """

        Parameters
        ----------
        in_features : int
            input dimension
        out_features : int
            output dimension
        bias : bool, optional
            use a bias vector, by default True
        device : Optional[str], optional
            device on which the graph should be defined (cpu or cuda), by default None
        dtype : _type_, optional
            dtype to be used in computations, by default None
        outputs_names : Optional[Sequence[str]], optional
            sequence of output names, by default None
        """
        super().__init__(in_features, out_features, bias, device, dtype)

        self.outputs_names = outputs_names

        self.subweight = None
        self.subbias = None

        self.eval()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Evaluates the linear layer, restricted or not depending of past settings.

        Parameters
        ----------
        x : torch.Tensor
            input tensor of shape (?, ``in_features``)

        Returns
        -------
        torch.Tensor
            output tensor. This tensor has shape (?, ``out_features``) if the full output set is considered, and with less outputs features in case of restricted output.
        """
        if self.training:
            return nn.functional.linear(x, self.weight, self.bias)
        return nn.functional.linear(x, self.subweight, self.subbias)

    def restrict_to_output_subset(self, indices: Sequence[int]) -> None:
        """Restricts the output to an output subset.

        Parameters
        ----------
        indices : Sequence[int]
            index subset to predict.

        Raises
        ------
        PermissionError
            the restriction cannot be applied in train mode.
            To apply the restriction, first turn the model to ``.eval()`` mode.
        """
        if self.training:
            raise PermissionError(
                "You're not able to restrict the outputs when Module mode is train"
            )
        self.subweight = self.weight.data[indices, :]
        if self.bias is not None:
            self.subbias = self.bias.data[indices]

    def train(self, mode: bool = True) -> "RestrictableLinear":
        r"""Sets the object in train mode if ``mode is True`` and in eval mode else.

        Parameters
        ----------
        mode : bool, optional
            Whether the layer is to be set in train mode (``True``) or eval mode (``False``), by default ``True``.

        Returns
        -------
        RestrictableLinear
            The updated object.
        """
        super().train(mode)
        if mode:
            self.subweight = None
            self.subbias = None
        else:
            self.subweight = self.weight.data
            self.subbias = None if self.bias is None else self.bias.data

    def __str__(self) -> str:
        return f"Restrictable layer ({self.in_features} input features, {self.out_features} output features)"



from enum import Enum

import pandas as pd

from .operator import Operator

__all__ = ["NormTypes", "Normalizer", "InverseNormalizer"]


class NormTypes(Enum):
    """Types of normalization"""

    NONE = 0
    """No normalization."""

    MEAN0 = 1
    """Center the columns, i.e., set their means to 0."""

    STD1 = 2
    """Reduce the columns, i.e., set their variances to 1."""

    MEAN0STD1 = 3
    """Center and reduce the columns, i.e., set their means to 0 and their variances to 1."""

    MIN0MAX1 = 4
    """Apply a MinMax normalization, i.e., set the minimum value of each column to 0 and the maximum to 1."""

    MIN1MAX1 = 5
    """Apply an alternative MinMax normalization, i.e., set the minimum value of each column to -1 and the maximum to 1."""


class Normalizer(Operator):
    r"""Specific operator that applies a specified normalization of the dataset."""

    def __init__(self, df: pd.DataFrame, norm_type: NormTypes):
        """

        Parameters
        ----------
        df : pd.DataFrame
            dataset to be normalized.
        norm_type : NormTypes
            type of normalization to be applied.
        """
        self.norm_type = norm_type

        self.stats = {
            "mean": df.mean().values,
            "std": df.std().values,
            "min": df.min().values,
            "max": df.max().values,
        }

    def __call__(self, t):
        """applies the specified normalization type to a set of values ``t``.

        Parameters
        ----------
        t : numpy.ndarray or pd.DataFrame
            set of values to normalize (typically one column of a dataset).

        Returns
        -------
        numpy.ndarray or pd.DataFrame
            normalized set of values.

        Raises
        ------
        RuntimeError
            The ``norm_type`` attribute is not a valid element of NormTypes.
        """
        if self.norm_type is NormTypes.NONE:
            return t
        elif self.norm_type is NormTypes.MEAN0:
            return t - self.stats["mean"]
        elif self.norm_type is NormTypes.STD1:
            return t / self.stats["std"]
        elif self.norm_type is NormTypes.MEAN0STD1:
            return (t - self.stats["mean"]) / self.stats["std"]
        elif self.norm_type is NormTypes.MIN0MAX1:
            return (t - self.stats["min"]) / (self.stats["max"] - self.stats["min"])
        elif self.norm_type is NormTypes.MIN1MAX1:
            return (
                2 * (t - self.stats["min"]) / (self.stats["max"] - self.stats["min"])
                - 1
            )
        else:
            raise RuntimeError("Should never been here")

    def __str__(self) -> str:
        return f"Normalizer: {self.norm_type}"


class InverseNormalizer(Operator):
    r"""Specific operator that reverts a specified normalization of the dataset."""

    def __init__(self, df: pd.DataFrame, norm_type: NormTypes):
        """

        Parameters
        ----------
        df : pd.DataFrame
            dataset to be normalized.
        norm_type : NormTypes
            type of normalization to revert.
        """
        self.norm_type = norm_type

        self.stats = {
            "mean": df.mean().values,
            "std": df.std().values,
            "min": df.min().values,
            "max": df.max().values,
        }

    def __call__(self, t):
        """reverts the specified normalization type to a set of normalized values ``t``.

        Parameters
        ----------
        t : numpy.ndarray or pd.DataFrame
            normalized set of values (typically one column of a dataset).

        Returns
        -------
        numpy.ndarray or pd.DataFrame
            unnormalized set of values

        Raises
        ------
        RuntimeError
            The ``norm_type`` attribute is not a valid element of NormTypes.
        """
        if self.norm_type is NormTypes.NONE:
            return t
        elif self.norm_type is NormTypes.MEAN0:
            return t + self.stats["mean"]
        elif self.norm_type is NormTypes.STD1:
            return t * self.stats["std"]
        elif self.norm_type is NormTypes.MEAN0STD1:
            return t * self.stats["std"] + self.stats["mean"]
        elif self.norm_type is NormTypes.MIN0MAX1:
            return t * (self.stats["max"] - self.stats["min"]) + self.stats["min"]
        elif self.norm_type is NormTypes.MIN1MAX1:
            return (t + 1) * (self.stats["max"] - self.stats["min"]) / 2 + self.stats[
                "min"
            ]
        else:
            raise RuntimeError("Should never been here")

    def __str__(self) -> str:
        return f"Inverse normalizer: {self.norm_type}"



import itertools as itt
from math import comb
from typing import Union

import torch
from numpy import ndarray
from torch import nn

__all__ = ["PolynomialExpansion"]


class PolynomialExpansion(nn.Module):
    r"""
    Polynomial expansion layer. For instance, for ``n_features=3`` and ``order=2``,


    .. math::

       \mathrm{poly}((x_1,\, x_2,\,x_3)) = (x_1,\,x_2,\,x_3,\,x_1^2,\,x_1x_2,\,x_1x_3,\,x_2^2,\,x_2x_3,\,x_3^2)

    """

    def __init__(
        self,
        n_features: int,
        order: int,
        device: str = "cpu",
    ):
        r"""

        Parameters
        ----------
        n_features : int
            Number of input features.
        order : int
            maximum degree of the polynomial expansion.
        device : str, optional
            Device to use, by default "cpu".
        """
        super().__init__()

        self.n_features = n_features
        self.order = order
        self.device = device

        # Expanded features
        self.n_expanded_features = PolynomialExpansion.expanded_features(
            order, n_features
        )

        # Mask creation
        self._mask = nn.Parameter(
            PolynomialExpansion._create_mask(order, n_features),
            requires_grad=False,
        )

        # Standardization
        self.count = nn.Parameter(torch.tensor(0, dtype=int), requires_grad=False)
        self.means = nn.Parameter(
            torch.zeros(self.n_expanded_features), requires_grad=False
        )
        self.squares = nn.Parameter(
            torch.ones(self.n_expanded_features), requires_grad=False
        )
        self.stds = nn.Parameter(
            torch.ones(self.n_expanded_features), requires_grad=False
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Applies the polynomial expansion.

        Parameters
        ----------
        x : Tensor
            Input tensor of shape (?, ``n_features``).

        Returns
        -------
        Tensor
            Output tensor of shape (?, ``expanded_features(order, n_features)``).
        """
        is1d = x.ndim == 1
        if is1d:
            x = x.unsqueeze(0)
        y = torch.cat(
            (torch.ones(x.shape[:-1], device=self.device).unsqueeze(-1), x), dim=-1
        )  # Add an input equal to 1
        m = y.clone()
        for _ in range(self.order - 1):
            y = y.unsqueeze(-1)
            m = m.unsqueeze(x.ndim - 1) * y
        y = m.reshape(*x.size()[:-1], -1).matmul(self._mask)
        y = (y - self.means) / self.stds
        if is1d:
            y = y.squeeze(0)
        return y

    @staticmethod
    def _create_mask(order: int, n_features: int) -> torch.Tensor:
        """Generates the mask used to perform the polynomial expansion.

        Parameters
        ----------
        order : int
            order of the polynomial expansion.
        n_features : int
            dimension of the input vector.

        Returns
        -------
        mask_mat: torch.Tensor
            binary mask
        """
        # Build hypercube
        mask = torch.ones(order * (n_features + 1,), dtype=bool)
        for coords in itt.product(*(range(n_features + 1) for _ in range(order))):
            # Select only the upper part of the tensor
            for k in range(order - 1):
                if coords[k + 1] < coords[k]:
                    mask[coords] = False
                    break

        # Flatten the hypercube
        mask_cube = mask.flatten()

        # Overlook 0 order expansion
        mask_cube[0] = False

        # Build the selection matrix
        n_rows, n_cols = mask_cube.numel(), mask_cube.sum()
        mask_mat = torch.zeros((n_rows, n_cols))
        idx = torch.arange(n_rows)[mask_cube]
        for i in range(n_cols):
            mask_mat[idx[i], i] = 1.0

        return mask_mat

    @staticmethod
    def expanded_features(order: int, n_features: int) -> int:
        """Returns the number of augmented polynomial features of order lower or equal to ``order`` and of ``n_features`` variables.


        Parameters
        ----------
        order : int
            maximum degree of the polynomial expansion.
        n_features : int
            Number of input features.

        Returns
        -------
        int
            number of augmented polynomial features of order lower or equal to ``order`` and of ``n_features`` variables.
        """
        return sum([comb(n_features + d - 1, d) for d in range(1, order + 1)])

    def update_standardization(
        self, x: Union[torch.Tensor, ndarray], reset: bool = False
    ) -> None:
        """
        This optional operation standardizes the output of the layer.
        Even in case of standardized inputs, the output of this layer are in general not standardized because of the polynomial transformations.
        This fonction can be called multiple times for different batches.
        To ensure correct moment calculation, entries must pass this functino only once.

        Parameters
        ----------
        x : torch.Tensor
            Input batch of shape (?, ``n_features``).
        reset: bool, optional
            If ``True``, reset moment calculation.
        """
        if isinstance(x, ndarray):
            x = torch.from_numpy(x)
        prev_mode = self.training
        self.eval()
        y = self.forward(x) * self.stds + self.means
        self.train(mode=prev_mode)

        y = y.reshape(-1, y.size(-1))

        if reset:
            self.count *= 0

        self.count += y.size(0)

        self.means *= 1 - y.size(0) / self.count
        self.means += y.size(0) / self.count * y.mean(dim=0)
        self.squares *= 1 - y.size(0) / self.count
        self.squares += y.size(0) / self.count * y.square().mean(dim=0)

        self.stds *= 0
        self.stds += torch.sqrt(self.squares - self.means**2)

    def __str__(self) -> str:
        return f"Polynomial expansion ({self.n_features} features, order {self.order})"



from .batch_scheduler import *
from .loss_functions import *
from .network_learning import *
from .regression_metrics import *

__all__ = []
__all__.extend(batch_scheduler.__all__)
__all__.extend(loss_functions.__all__)
__all__.extend(network_learning.__all__)
__all__.extend(regression_metrics.__all__)



from typing import Iterable, Optional, Sequence, Union

from torch import Tensor, nn

from ..layers import RestrictableLinear
from ..operators import Operator
from .neural_network import NeuralNetwork

__all__ = ["FullyConnected"]


class FullyConnected(NeuralNetwork):
    r"""Standard fully connected neural network."""

    def __init__(
        self,
        layers_sizes: Iterable[int],
        activation: nn.Module,
        batch_norm: bool = False,
        inputs_names: Optional[Sequence[str]] = None,
        outputs_names: Optional[Sequence[str]] = None,
        inputs_transformer: Optional[Operator] = None,
        outputs_transformer: Optional[Operator] = None,
        device: Optional[str] = None,
        last_restrictable: bool = True,
    ):
        """

        Parameters
        ----------
        layers_sizes : Iterable[int]
            list of number of neurons per layer. The first value corresponds to the dimension of the input layer, and the last value to the dimension of the output layer.
        activation : nn.Module
            activation function.
        batch_norm : bool, optional
            whether to use batch normalization during training, by default ``False``.
        inputs_names : Optional[Sequence[str]], optional
            List of inputs names. None if the names have not been specified. By default None.
        outputs_names : Optional[Sequence[str]], optional
            List of outputs names. None if the names have not been specified. By default None.
        inputs_transformer : Optional[Operator], optional
            Transformation applied to the inputs before processing, by default None.
        outputs_transformer : Optional[Operator], optional
            Transformation applied to the outputs after processing, by default None.
        device : Optional[str], optional
            Device used ("cpu" or "cuda"), by default None (corresponds to "cpu").
        last_restrictable : bool, optional
            whether the last layer is to be a RestrictableLinear layer, by default ``True``.
        """
        super().__init__(
            layers_sizes[0],
            layers_sizes[-1],
            inputs_names=inputs_names,
            outputs_names=outputs_names,
            inputs_transformer=inputs_transformer,
            outputs_transformer=outputs_transformer,
            device=device,
        )

        self.layers_sizes = layers_sizes

        self.activation = activation
        self.batch_norm = batch_norm

        self.last_restrictable = last_restrictable

        self.layers = nn.ModuleList()
        for k in range(len(layers_sizes) - 2):
            if batch_norm:
                self.layers.append(
                    nn.Sequential(
                        nn.Linear(
                            self.layers_sizes[k],
                            self.layers_sizes[k + 1],
                            bias=False,
                            device=self.device,
                        ),
                        nn.BatchNorm1d(
                            self.layers_sizes[k + 1],
                            device=self.device,
                        ),
                    )
                )
            else:
                self.layers.append(
                    nn.Linear(
                        self.layers_sizes[k],
                        self.layers_sizes[k + 1],
                        bias=True,
                        device=self.device,
                    )
                )

        if last_restrictable:
            self.output_layer = RestrictableLinear(
                layers_sizes[-2],
                layers_sizes[-1],
                outputs_names=outputs_names,
                device=self.device,
            )
        else:
            self.output_layer = nn.Linear(
                layers_sizes[-2],
                layers_sizes[-1],
                device=self.device,
            )

    def forward(self, x: Tensor) -> Tensor:
        y_hat = x.clone()

        is1d = y_hat.ndim == 1
        if is1d:
            y_hat = y_hat.unsqueeze(0)

        for layer in self.layers:
            y_hat = layer(y_hat)
            y_hat = self.activation(y_hat)

        y_hat = self.output_layer(y_hat)

        if not self.last_restrictable:
            y_hat = y_hat[..., self.current_output_subset_indices]
        if is1d:
            y_hat = y_hat.squeeze(0)

        return y_hat

    def restrict_to_output_subset(
        self, output_subset: Optional[Union[Sequence[str], Sequence[int]]]
    ) -> None:
        super().restrict_to_output_subset(output_subset)
        if self.last_restrictable:
            self.output_layer.restrict_to_output_subset(
                self.current_output_subset_indices
            )



from math import log

import torch

__all__ = [
    "error_factor_log",
    "error_factor_lin",
    "relative_error",
    "squared_error",
    "absolute_error",
]

LOG10 = log(10)


def error_factor_log(y_hat: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    err = torch.exp(LOG10 * torch.abs(y_hat.detach() - y))
    return 100 * (err - 1)


def error_factor_lin(y_hat: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    err = torch.maximum(torch.abs(y_hat / y), torch.abs(y / y_hat))
    return 100 * (err - 1)


def relative_error(y_hat: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    return 100 * (y_hat - y) / y


def squared_error(y_hat: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    return (y_hat - y).square()


def absolute_error(y_hat: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    return (y_hat - y).abs()



from .dataset import *
from .layers import *
from .learning import *
from .networks import *
from .operators import *

__all__ = []
__all__.extend(dataset.__all__)
__all__.extend(layers.__all__)
__all__.extend(networks.__all__)
__all__.extend(learning.__all__)
__all__.extend(operators.__all__)



from abc import ABC, abstractmethod
from typing import Callable, Optional, Union

import torch
from torch import nn

__all__ = ["AdditionalModule", "AdditionalModuleFromExisting"]


class AdditionalModule(nn.Module, ABC):
    r"""
    Additional module.
    """

    def __init__(
        self,
        input_features: Optional[int],
        output_features: Optional[int],
        device: str = "cpu",
    ):
        r"""

        Parameters
        ----------
        input_features: int
            Number of input features.
        output_features: int
            Number of output features.
        device: str
            Device to use, by default "cpu".
        """
        super().__init__()

        self.input_features = input_features
        self.output_features = output_features
        self.device = device

    @abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        r"""
        Evaluates the associated pytorch function.

        Parameters
        ----------
        x : Tensor
            Input tensor of shape (?, ``input_features``).

        Returns
        -------
        Tensor
            Output tensor of shape (?, ``output_features``).
        """

    def __str__(self) -> str:
        return f"Additional module ({self.input_features} input features, {self.output_features} output features)"


class AdditionalModuleFromExisting(AdditionalModule):
    r"""
    Additional module build from an existing Torch module or function.
    Avoid overriding forward method.
    """

    def __init__(
        self,
        input_features: Optional[int],
        output_features: Optional[int],
        operation: Union[nn.Module, Callable],
        device: str = "cpu",
    ):
        r"""

        Parameters
        ----------
        input_features: int
            Number of input features.
        output_features: int
            Number of output features.
        operation: Module | function
            Operation to apply (Torch module or function).
        device: str
            Device to use, by default "cpu".
        """
        super().__init__(input_features, output_features, device=device)

        self.operation = operation

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        r"""
        Evaluates the associated pytorch function.

        Parameters
        ----------
        x : Tensor
            Input tensor of shape (?, ``input_features``).

        Returns
        -------
        Tensor
            Output tensor of shape (?, ``output_features``).
        """
        if isinstance(self.operation, nn.Module):
            return self.operation.forward(
                x
            )  # To prevent cases where __call__ is overriden and may be different from forward
        return self.operation(x)

    def __str__(self) -> str:
        return f"Additional module from an existing Torch operator ({self.input_features} input features, {self.output_features} output features, module: {self.module})"



from abc import ABC, abstractmethod

__all__ = [
    "BatchScheduler",
    "LinearBatchScheduler",
    "ExponentialBatchScheduler",
]


class BatchScheduler(ABC):
    r"""Abstract class for schedulers of the batch size :math:`b(t)` with respect to the ``epoch`` attribute, denoted :math:`t` in math equations."""

    def __init__(self):
        self.epoch = 1

    def step(self) -> None:
        r"""Increments the ``epoch`` attribute: :math:`t \leftarrow t+1`."""
        self.epoch += 1

    def set_epoch(self, epoch: int) -> None:
        """Set the ``epoch`` attribute to a new value.

        Parameters
        ----------
        epoch : int
            new value for the ``epoch`` attribute.
        """
        self.epoch = epoch

    @abstractmethod
    def get_batch_size(self) -> int:
        """Returns the batch size :math:`b(t)` for the next epoch depending on the number of already run epochs :math:`t`.

        Returns
        -------
        int
            batch size :math:`b(t)`.
        """
        pass


class ConstantBatch:
    r"""Simplest scheduler, that always returns a predefined constant :math:`c`, i.e., :math:`b(t) = c` for all :math:`t`"""
    # batch_size: int

    def __init__(self, batch_size: int):
        """

        Parameters
        ----------
        batch_size : int
            batch size to consider during the training.
        """
        super().__init__()
        self.batch_size = batch_size

    def get_batch_size(self) -> int:
        return self.get_batch_size


class LinearBatchScheduler(BatchScheduler):
    r"""Scheduler based on a linear interpolation between an initial batch size :math:`b_{i}` and a final batch size :math:`b_{f}` for a total number of epochs :math:`t_{f}`, i.e.,  for :math:`0 \leq t \leq t_{f}`,


    .. math::

       b(t) = \frac{t}{t_{f}} b_{i} + \frac{t_{f} - t}{t_{f}} b_{f}

    """

    # start: int
    # stop: int
    # n_epochs: int

    def __init__(self, start: int, stop: int, n_epochs: int):
        r"""

        Parameters
        ----------
        start : int
            starting value for the batch size :math:`b_{i}`.
        stop : int
            final value for the batch size :math:`b_{f}`.
        n_epochs : int
            total number of epochs considered for training :math:`t_{f}`.
        """
        super().__init__()
        self.start = start
        self.stop = stop
        self.n_epochs = n_epochs

    def get_batch_size(self) -> int:
        slope = (self.stop - self.start) / self.n_epochs
        return round((self.epoch - 1) * slope + self.start)


class ExponentialBatchScheduler(BatchScheduler):
    r"""Scheduler based on an exponential interpolation between an initial batch size :math:`b_{i}` and a final batch size :math:`b_{f}` for a total number of epochs :math:`t_{f}`, i.e., for :math:`0 \leq t \leq t_{f}`,


    .. math::

       b(t) = \left( \frac{b_{f}}{b_{i}} \right)^{t / t_{f}} b_{i}

    """
    # start: int
    # stop: int
    # n_epochs: int

    def __init__(self, start: int, stop: int, n_epochs: int):
        r"""

        Parameters
        ----------
        start : int
            starting value for the batch size :math:`b_{i}`.
        stop : int
            final value for the batch size :math:`b_{f}`.
        n_epochs : int
            total number of epochs considered for training :math:`t_{f}`.
        """
        super().__init__()
        self.start = start
        self.stop = stop
        self.n_epochs = n_epochs

    def get_batch_size(self) -> int:
        factor = (self.stop / self.start) ** (1 / self.n_epochs)
        return round(factor ** (self.epoch - 1) * self.start)



import os
import pickle
from typing import (
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)

import numpy as np
import pandas as pd
from torch import Tensor, from_numpy
from torch.utils.data import Dataset

__all__ = ["RegressionDataset", "RegressionSubset"]


class RegressionDataset(Dataset):
    """Dataset dedicated to regression."""

    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
        inputs_names: Optional[List[str]] = None,
        outputs_names: Optional[List[str]] = None,
    ):
        r"""

        Parameters
        ----------
        x : numpy.ndarray
            Array containing the input features of the regression model.
            Its shape is considered to be :math:`N \times I` where :math:`N` is the number of entries and :math:`I` the number of input features.
        y : numpy.ndarray
            Array containing the output features of the regression model.
            Its shape is considered to be :math:`N \times O` where :math:`N` is the number of entries and :math:`O` the number of output features.
        inputs_names : Optional[List[str]], optional
            list of the names of the input features, by default None
        outputs_names : Optional[List[str]], optional
            list of the names of the output features, by default None

        Raises
        ------
        ValueError
            ``x`` and ``y`` must have the same number of rows :math:`N`.
        ValueError
            ``x`` and ``inputs_names`` must have the same number of features :math:`I`.
        ValueError
            ``y`` and ``outputs_names`` must have the same number of features :math:`O`.
        """
        super().__init__()

        if x.shape[0] != y.shape[0]:
            raise ValueError("x and y must have the same number of rows.")

        self._x: Tensor = from_numpy(x).float()
        self._y: Tensor = from_numpy(y).float()

        if inputs_names is not None and len(inputs_names) != x.shape[1]:
            raise ValueError("x and inputs_names must have the same number of features")
        if outputs_names is not None and len(outputs_names) != y.shape[1]:
            raise ValueError(
                "y and outputs_names must have the same number of features"
            )

        self._inputs_names: Optional[List[str]] = inputs_names
        self._outputs_names: Optional[List[str]] = outputs_names

    def __len__(self) -> int:
        """
        Returns the number of entries :math:`N` in the dataset.

        Returns
        -------
        int
            Number of entries :math:`N`.
        """
        return self._x.size(0)

    def __getitem__(self, idx) -> Tuple[Tensor, Tensor]:
        """
        Returns the entries corresponding to the index set ``idx``.

        Parameters
        ----------
        idx : Any
            Indices of entries to return.

        Returns
        -------
        tuple of torch.Tensor
            Input and output entries.
        """
        return self._x[idx], self._y[idx]

    @property
    def x(self) -> Tensor:
        """
        Input tensor.
        """
        return self._x

    @property
    def y(self) -> Tensor:
        """
        Output tensor.
        """
        return self._y

    @property
    def n_inputs(self) -> int:
        """
        Number of input features :math:`I`.
        """
        return self.x.size(1)

    @property
    def n_outputs(self) -> int:
        """
        Number of output features :math:`O`.
        """
        return self.y.size(1)

    @property
    def inputs_names(self) -> int:
        """
        List of the names of the input features.
        """
        return self._inputs_names

    @property
    def outputs_names(self) -> int:
        """
        List of the names of the output features.
        """
        return self._outputs_names

    def inputs_size(self) -> int:
        """
        Returns the number of floating point values in ``x``.
        """
        return self.x.numel()

    def outputs_size(self) -> int:
        """
        Returns the number of floating point values in ``y``.
        """
        return self.y.numel()

    def has_nan(self) -> Tuple[bool, bool]:
        """
        Returns a tuple of two boolean.
        The first one is ``True`` if the input features contain at least one NaN, else ``False``.
        The second one is ``True`` if the output features contain at least one NaN, else ``False``.

        Returns
        -------
        tuple of bool
            Evaluate the presence of NaN in the input and output sets.
        """
        return self.x.isnan().any().item(), self.y.isnan().any().item()

    def has_nonfinite(self) -> Tuple[bool, bool]:
        """
        Returns a tuple of two boolean.
        The first one is ``True`` if the input features contain at least one non finite value (including NaNs), else ``False``.
        The second one is ``True`` if the output features contain at least one non finite value (including NaNs), else ``False``.

        Returns
        -------
        tuple of bool
            Evaluate the presence of non finite values in the input and output sets.
        """
        return not self.x.isfinite().all().item(), not self.y.isfinite().all().item()

    def apply_transf(
        self,
        x_op: Optional[Callable[[np.ndarray], np.ndarray]],
        y_op: Optional[Callable[[np.ndarray], np.ndarray]],
    ) -> "RegressionDataset":
        """
        Apply an operator to ``x`` and ``y``. A new dataset is returned so the operators should not use in-place operations.

        Parameters
        ----------
        x_op : Callable[[numpy.ndarray], np.ndarray]
            Operator to apply on the input features.
        y_op : Callable[[numpy.ndarray], np.ndarray]
            Operator to apply on the output features.

        Returns
        -------
        RegressionDataset
            New dataset with transformed values.
        """
        if x_op is None:
            x_op = lambda t: t
        if y_op is None:
            y_op = lambda t: t
        return type(self)(
            x_op(self.x.numpy()),
            y_op(self.y.numpy()),
            self._inputs_names,
            self._outputs_names,
        )

    @overload
    def getall(self, numpy: Literal[True]) -> Tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def getall(self, numpy: Literal[False]) -> Tuple[Tensor, Tensor]:
        ...

    def getall(self, numpy: bool = False) -> Tuple[Union[np.ndarray, Tensor], ...]:
        """
        Returns all the dataset in numpy.ndarray or torch.Tensor depending on the value of the ``numpy`` parameter.

        Parameters
        ----------
        numpy : bool, optional
            If ``numpy==True``, the returned object will be numpy arrays.
            Else, they will be torch tensors.

        Returns
        -------
        tuple of torch.Tensor or numpy.ndarray
            Inputs and outputs sets.
        """
        x, y = self[list(range(len(self)))]
        if numpy:
            return x.numpy(), y.numpy()
        return x, y

    @staticmethod
    def from_pandas(df_x: pd.DataFrame, df_y: pd.DataFrame) -> "RegressionDataset":
        r"""Converts two pandas DataFrames to a RegressionDataset object.

        Parameters
        ----------
        df_x : pd.DataFrame
            DataFrame of the inputs. This DataFrame should contain :math:`N` rows, i.e., number of entries, and :math:`I` columns, i.e., features.
        df_y : pd.DataFrame
            DataFrame of the outputs. This DataFrame should contain :math:`N` rows, i.e., number of entries, and :math:`O` columns, i.e., features.

        Returns
        -------
        RegressionDataset
            associated RegressionDataset object.
            The ``x`` attribute is set to values in the ``df_x`` DataFrame, and the ``input_names`` attribute to its column names.
            The ``y`` attribute is set to values in the ``df_y`` DataFrame, and the ``output_names`` attribute to its column names.
        """
        return RegressionDataset(
            df_x.values,
            df_y.values,
            df_x.columns.to_list(),
            df_y.columns.to_list(),
        )

    def to_pandas(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Converts the dataset to two pandas DataFrames.

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame]
            DataFrames of the input ``x`` and output ``y``, respectively. The columns are names with the ``input_names`` and ``output_names``, respectively, if they are not ``None``.
        """
        return pd.DataFrame(self.x, columns=self._inputs_names), pd.DataFrame(
            self.y, columns=self._outputs_names
        )

    def join(
        self: "RegressionDataset", other: "RegressionDataset"
    ) -> "RegressionDataset":
        """
        Returns the union of two datasets. Data are copied.

        Parameters
        ----------
        other : RegressionDataset
            Other dataset to join with.

        Returns
        -------
        type
            New dataset constructed as the union of the two datasets.
        """
        x1, y1 = self.getall(numpy=True)
        x2, y2 = other.getall(numpy=True)
        x = np.concatenate((x1, x2), axis=0)
        y = np.concatenate((y1, y2), axis=0)
        return RegressionDataset(x, y)

    def substract(
        self: "RegressionDataset", other: "RegressionSubset"
    ) -> "RegressionSubset":
        """
        Returns the subtraction of two datasets. Data are copied.

        Description.

        Parameters
        ----------
        other : RegressionSubset
            Subset of ``self``.

        Returns
        -------

        RegressionSubset
            New subset of ``self`` containing all values that were not in `other`.
        """
        if not other.issubsetof(self):
            raise ValueError(
                "set2 is not a subset of set1 so it cannot be substracted."
            )
        new_indices = [i for i in range(len(self)) if i not in other._indices]
        # Algo can be improved.
        return RegressionSubset(self, new_indices)

    def stats(self) -> Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray]]:
        """Provides a few statistics on the dataset (the mean, the standard deviation, the min and the max for each column).

        Returns
        -------
        Tuple[ Dict[str, np.ndarray], Dict[str, np.ndarray] ]
            Tuple of dictionaries, each containing the mean, the standard deviation, the min and the max for each column. The first dictionary corresponds to the input x and the second to the output y.
        """
        return {
            "mean": self.x.mean(axis=0).numpy(),
            "std": self.x.std(axis=0).numpy(),
            "min": self.x.min(axis=0).numpy(),
            "max": self.x.max(axis=0).numpy(),
        }, {
            "mean": self.y.mean(axis=0).numpy(),
            "std": self.y.std(axis=0).numpy(),
            "min": self.y.min(axis=0).numpy(),
            "max": self.y.max(axis=0).numpy(),
        }

    def save(self, filename: str, path: Optional[str] = None) -> None:
        """saves the dataset to a pickle file.

        Parameters
        ----------
        filename : str
            name of the file to be created.
        path : Optional[str], optional
            path to the file to be created, by default None
        """
        if path is not None:
            filename = os.path.join(path, filename)
        filename = os.path.splitext(filename)[0]
        with open(f"{filename}.pkl", "wb") as file:
            pickle.dump(self, file)

    @staticmethod
    def load(filename: str, path: Optional[str] = None) -> "RegressionDataset":
        """loads a regression dataset from a pickle file.

        Parameters
        ----------
        filename : str
            name of the file to be read.
        path : Optional[str], optional
            path to the file to be read, by default None.

        Returns
        -------
        RegressionDataset
            loaded regression dataset.
        """
        if path is not None:
            filename = os.path.join(path, filename)
        filename = os.path.splitext(filename)[0]
        with open(f"{filename}.pkl", "rb") as file:
            dataset = pickle.load(file)
        return dataset


class RegressionSubset(RegressionDataset):
    """Subset of RegressionDataset."""

    def __init__(self, dataset: RegressionDataset, indices: Sequence[int]):
        """

        Attributes
        ----------
        dataset : RegressionDataset
            Dataset from which entries are extracted.
        indices : Sequence of int
            Indices of entries to extract.
        """
        self._dataset: RegressionDataset = dataset
        self._indices: Sequence[int] = indices

        self._inputs_names = dataset.inputs_names
        self._outputs_names = dataset.outputs_names

    def __len__(self) -> int:
        """
        Returns the number of entries in the dataset.

        Returns
        -------
        int
            Number of entries.
        """
        return len(self._indices)

    def __getitem__(self, idx) -> Tuple[Tensor, Tensor]:
        """
        Returns the entries of indice(s) idx.

        Parameters
        ----------
        idx : Any
            Indices of entries to return.

        Returns
        -------
        tuple of torch.Tensor
            Input and output entries.
        """
        return self._dataset[self._indices[idx]]

    @property
    def x(self) -> Tensor:
        """
        Input tensor.
        """
        return self._dataset._x[self._indices]

    @property
    def y(self) -> Tensor:
        """
        Output tensor.
        """
        return self._dataset._y[self._indices]

    def issubsetof(self, dataset: RegressionDataset) -> bool:
        """
        Returns ``True`` of ``self`` is a subset of ``dataset``.

        Parameters
        ----------
        dataset : RegressionDataset
            Dataset of which we want to know if ``self`` is a subset.

        Returns
        -------
        bool
            ``True`` of ``self`` is a subset of ``dataset`` else ``False``.
        """
        # TODO maybe a better solution is suitable
        return dataset == self._dataset



from .restrictable_linear import *
from .polynomial_expansion import *
from .additional_module import *

__all__ = []

__all__.extend(restrictable_linear.__all__)
__all__.extend(polynomial_expansion.__all__)
__all__.extend(additional_module.__all__)



import json
import os
import pickle
import shutil
import time
from abc import ABC, abstractmethod
from collections import OrderedDict
from inspect import signature
from typing import List, Literal, Optional, Tuple, Union, overload

import numpy as np
import torch
from torch import nn

from ..operators import Operator

__all__ = ["NeuralNetwork"]


class NeuralNetwork(nn.Module, ABC):
    r"""Neural network abstract class."""

    def __init__(
        self,
        input_features: int,
        output_features: int,
        inputs_names: Optional[List[str]] = None,
        outputs_names: Optional[List[str]] = None,
        inputs_transformer: Optional[Operator] = None,
        outputs_transformer: Optional[Operator] = None,
        device: Optional[str] = None,
    ):
        r"""

        Parameters
        ----------
        input_features : int
            Dimension of input vector.
        output_features : int
            Dimension of output vector.
        inputs_names: Optional[List[str]], optional
            List of inputs names. None if the names have not been specified.
        outputs_names: Optional[List[str]], optional
            List of outputs names. None if the names have not been specified.
        inputs_transformer : Optional[Operator], optional
            Transformation applied to the inputs before processing.
        outputs_transformer: Optional[Operator], optional
            Transformation applied to the outputs after processing.
        device : Optional[str], optional
            Device used ("cpu" or "cuda"), by default None (corresponds to "cpu").

        Raises
        ------
        ValueError
            The length of ``inputs_names`` should equal ``input_features``.
        ValueError
            The length of ``outputs_names`` should equal ``output_features``.
        ValueError
            The list ``inputs_names`` must not have duplicates.
        ValueError
            The list ``outputs_names`` must not have duplicates.
        """
        super().__init__()
        self.input_features = input_features
        self.output_features = output_features

        if inputs_names is not None and len(inputs_names) != input_features:
            raise ValueError(
                f"inputs_names of length {len(inputs_names)} is incompatible with {input_features} input_features"
            )
        if outputs_names is not None and len(outputs_names) != output_features:
            raise ValueError(
                f"outputs_names of length {len(outputs_names)} is incompatible with {output_features} outputs_features"
            )

        if inputs_names is not None and len(set(inputs_names)) != len(inputs_names):
            raise ValueError("inputs_names has duplicates")
        if outputs_names is not None and len(set(outputs_names)) != len(outputs_names):
            raise ValueError("outputs_names has duplicates")

        self.inputs_names = inputs_names
        self.outputs_names = outputs_names

        self.inputs_transformer = inputs_transformer
        self.outputs_transformer = outputs_transformer

        self.current_output_subset = outputs_names
        self.current_output_subset_indices = list(range(output_features))

        if device is None:
            self.device = "cuda" if torch.cuda.is_available() else "cpu"
        else:
            self.device = device

        self.eval()

    @abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        r"""Evaluates the neural network on an input ``x``.

        Parameters
        ----------
        x : torch.Tensor
            input tensor of shape (?, ``input_features``)

        Returns
        -------
        torch.Tensor
            output tensor of shape (?, ``output_features``)
        """
        pass

    def set_device(self, device: str) -> None:
        """
        Set the device to use.

        Parameters
        ----------
        device : str
            Device to use ("cpu" or "cuda").
        """
        assert device in ["cuda", "cpu"], f"device = {device}"
        self.device = device

    def evaluate(
        self,
        x: np.ndarray,
        transform_inputs: bool = False,
        transform_outputs: bool = False,
    ) -> np.ndarray:
        r"""Evaluate the network on a batch of NumPy inputs.

        Parameters
        ----------
        x : ndarray
            batch of inputs of shape (?, ``input_features``).
        transform_inputs : bool, optional
            whether the input ``x`` are to be pre-processed.
        transform_outputs : bool, optional
            whether the predictions ``y`` are to be post-processed.

        Returns
        -------
        ndarray
            batch of outputs of shape (?, ``output_features``).

        Raises
        ------
        ValueError
            The ``transform_inputs`` argument cannot be ``True`` when the ``inputs_transformer`` attribute is None.
        ValueError
            The ``transform_outputs`` argument cannot be ``True`` when the ``outputs_transformer`` attribute is None.
        """
        if transform_inputs:
            if self.inputs_transformer is None:
                raise ValueError(
                    "transform_inputs cannot be True when self.inputs_transformer is None"
                )
            x = self.inputs_transformer(x)

        p = next(self.parameters())
        x = torch.from_numpy(x).to(self.device)
        if x.dtype != p.dtype:
            x = x.type(p.dtype)
        with torch.no_grad():
            y = self.forward(x)
        y = y.detach().cpu().numpy()

        if transform_outputs:
            if self.outputs_transformer is None:
                raise ValueError(
                    "transform_outputs cannot be True when self.outputs_transformer is None"
                )
            y = self.outputs_transformer(y)

        return y

    @overload
    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        pass

    @overload
    def __call__(self, x: np.ndarray) -> np.ndarray:
        pass

    def __call__(
        self, x: Union[torch.Tensor, np.ndarray]
    ) -> Union[torch.Tensor, np.ndarray]:
        r"""Process a batch of NumPy ndarray or PyTorch Tensor inputs.

        Parameters
        ----------
        x : ndarray | Tensor
            batch of inputs of shape (?, ``input_features``).

        Returns
        -------
        ndarray | Tensor
            batch of outputs of shape (?, ``output_features``).
        """
        if isinstance(x, np.ndarray):
            return self.evaluate(x, False, False)
        else:
            with torch.no_grad():
                return self.forward(x)

    def train(self, mode: bool = True) -> "NeuralNetwork":
        r"""Set the current mode of the network (train or eval).

        Parameters
        ----------
        mode: bool, optional
            If ``True``, activate the training mode. If ``False``, activate the evaluation mode. Default: ``True``.

        Returns
        -------
        NeuralNetwork
            Instance of network.
        """
        super().train(mode)
        if mode:
            self.current_output_subset = self.outputs_names
            self.current_output_subset_indices = list(range(self.output_features))

    def restrict_to_output_subset(
        self,
        output_subset: Optional[Union[List[str], List[int]]] = None,
    ) -> None:
        r"""Restricts network outputs to those contained in ``output_subset``.

        Parameters
        ----------
        output_subset : List[str] | List[int] | None, optional
            Network outputs required. If None, no restriction is applied. Default: None.

        Raises
        ------
        PermissionError
            The outputs cannot be restricted when Module is in train mode.
        TypeError
            The ``output_subset`` argument must be a list or None.
        ValueError
            The ``output_subset`` list must not be empty.
        TypeError
            The ``output_subset`` argument must be a list of int or a list of str.
        """
        if self.training:
            raise PermissionError(
                "You're not able to restrict the outputs when Module mode is train"
            )

        if output_subset is None:
            self.current_output_subset = self.inputs_names
            self.current_output_subset_indices = list(range(self.output_features))
            return

        if not isinstance(output_subset, List):
            raise TypeError("output_subset must be a list or None")
        if len(output_subset) == 0:
            raise ValueError("output_subset must not be empty")

        if all(isinstance(x, int) for x in output_subset):
            self.current_output_subset = self._names_of_output_subset(output_subset)
            self.current_output_subset_indices = output_subset
        elif all(isinstance(x, str) for x in output_subset):
            self.current_output_subset = output_subset
            self.current_output_subset_indices = self._indices_of_output_subset(
                output_subset
            )
        else:
            raise TypeError("output_subset must be a list of int or a list of str")

    def _names_of_output_subset(self, output_subset: List[int]) -> List[str]:
        r"""Returns the names of outputs corresponding to the indices list ``output_subset``.

        Parameters
        ----------
        output_subset : List[int]
            Indices list.

        Returns
        -------
        List[str]
            Names of associated outputs.
        """
        if not isinstance(output_subset, (list, tuple)):
            raise TypeError("output_subset must be a list")
        if any(not isinstance(x, int) for x in output_subset):
            raise TypeError("output_subset must be a list of int")

        if not self._check_if_sublist(list(range(self.output_features)), output_subset):
            raise ValueError("input_subset is not a valid subset")

        return [self.outputs_names[k] for k in output_subset]

    def _indices_of_output_subset(self, output_subset: List[str]) -> List[int]:
        r"""Returns the indices of outputs corresponding to the names list ``output_subset``.

        Parameters
        ----------
        output_subset : List[str]
            Names list.

        Returns
        -------
        List[int]
            Indices of outputs.
        """
        if not isinstance(output_subset, (list, tuple)):
            raise TypeError("output_subset must be a list")
        if any(not isinstance(x, str) for x in output_subset):
            raise TypeError("output_subset must be a list of int")

        if self.outputs_names is None:
            raise TypeError(
                "output_subset cannot be a list of str when self.outputs_names is None"
            )
        if not self._check_if_sublist(self.outputs_names, output_subset):
            raise ValueError("output_subset is not a valid subset")

        return self._indices_of_sublist(self.outputs_names, output_subset)

    @staticmethod
    def _check_if_sublist(seq: List, subseq: List) -> bool:
        r"""Returns ``True`` if all elements of ``subseq`` are also elements of ``seq``.
        Else, returns ``False``.
        Duplicates in ``subseq`` do not affect the result.

        Parameters
        ----------
        seq : List
            Reference list.
        subseq : List
            List to be checked for inclusion in `seq`.

        Returns
        -------
        bool
            whether all elements of ``subseq`` are also elements of ``seq``.
        """
        return set(subseq) <= set(seq)

    @staticmethod
    def _indices_of_sublist(seq: List, subseq: List) -> List[int]:
        r"""Returns the indices in ``seq`` of the elements of ``subseq``. If ``subseq`` contains duplicates, the returned list will also contain duplicates.

        Parameters
        ----------
        seq : List
            Reference list.
        subseq : List
            List whose elements will be retrieved in ``seq``.

        Returns
        -------
        List[int]
            Indices of elements of ``subset`` in ``seq``.
        """
        if not NeuralNetwork._check_if_sublist(seq, subseq):
            raise ValueError("subseq is not a sublist of seq")
        index_dict = dict((value, idx) for idx, value in enumerate(seq))
        return [index_dict[value] for value in subseq]  # Remark: the result is ordered

    def count_parameters(self, learnable_only: bool = True) -> int:
        r"""Returns the number of parameters of the module.
        If ``learnable_only=True``, this function returns the number of parameters for which ``requires_grad=True``.
        If ``learnable_only=False``, it returns the number of parameters, independently to their ``requires_grad`` property.

        Parameters
        ----------
        learnable_only : bool, optional
            Indicates the the type of parameter to count. Defaults to True.

        Returns
        -------
        int
            Number of parameters.
        """
        self.train()
        if learnable_only:
            count = sum(p.numel() for p in self.parameters() if p.requires_grad)
        count = sum(p.numel() for p in self.parameters())
        self.eval()
        return count

    def count_bytes(
        self,
        learnable_only: bool = True,
        display: bool = False,
    ) -> Union[Tuple[int, Literal["B", "kB", "MB", "GB", "TB"]], str]:
        r"""Returns the number of parameters of the module.
        If ``learnable_only=True``, this function returns the number of parameters for which ``requires_grad=True``.
        If ``learnable_only=False``, it returns the number of parameters, independently to their ``requires_grad`` property.

        Parameters
        ----------
        learnable_only (bool, optional)
            Indicates the the type of parameter to count. Defaults to True.

        Returns
        -------
        int
            Number of parameters.
        str
            Unit ('B', 'kB', 'MB', 'GB', 'TB')
        """
        self.train()
        size = 0.0
        for p in self.parameters():
            if p.requires_grad or not learnable_only:
                size += (
                    p.numel() * p.element_size()
                )  # Does not take into consideration the Python object size which can be obtain using sys.getsizeof()
        self.eval()

        for v, u in [(1e0, "B"), (1e3, "kB"), (1e6, "MB"), (1e9, "GB"), (1e12, "TB")]:
            if size < 1e3 * v:
                if display:
                    return f"{size / v:.2f} {u}"
                else:
                    return (size / v, u)

    def time(self, n: int, repeat: int) -> Tuple[float, float, float]:
        """Estimates the evaluation time of the model for a batch of ``n`` inputs. Returns the average, min and max durations (in seconds) over ``repeat`` iterations.

        Parameters
        ----------
        n : int
            batch size.
        repeat : int
            number of evaluations.

        Returns
        -------
        Tuple[float, float, float]
            average, min and max durations (in seconds)

        Raises
        ------
        TypeError
            The ``n`` argument must be an integer.
        TypeError
            The ``repeat`` argument must be an integer.
        """
        if not isinstance(n, int):
            raise TypeError(f"n must be an integer, not {type(n)}")
        if not isinstance(repeat, int):
            raise TypeError(f"repeat must be an integer, not {type(repeat)}")
        times = []
        for it in range(repeat):
            x = torch.normal(0.0, 1.0, size=(n, self.input_features))
            tic = time.time()
            self.forward(x)
            toc = time.time()
            times.append(toc - tic)
        return sum(times) / repeat, min(times), max(times)

    def save(
        self,
        module_name: str,
        module_path: Optional[str] = None,
        overwrite: bool = True,
    ) -> None:
        r"""Saves the network for future use.

        Parameters
        ----------
        module_name: str
            Name of the directory in which the model will be saved.
        module_path: str | None
            Path to the previous directory.
        overwrite: bool
            If ``True``, the save can overwrite a previous backup of the same name. If ``False``, an error will be raised if such a backup exists.
        """
        if module_path is not None and not os.path.isdir(module_path):
            os.mkdir(module_path)

        if module_path is None:
            path = module_name
        else:
            path = os.path.join(module_path, module_name)

        if not os.path.exists(path):
            pass
        elif not overwrite:
            raise FileExistsError(f"{path} already exists")
        else:
            for _, __, f in os.walk(path):
                if not all(
                    os.path.isdir(file) or file.endswith((".json", ".pkl", ".pth"))
                    for file in f
                ):
                    raise ValueError(
                        f"{path} directory cannot be overwritten because it doesn't seem to be a NeuralNetwork save directory"
                    )
            shutil.rmtree(path)

        NeuralNetwork._recursive_save(self, path)

    @staticmethod
    def _recursive_save(
        module: nn.Module,
        path: str,
    ) -> None:
        r"""Make a recursive save of a PyTorch module. This makes it possible to deal with cases where some of the parameters of a network are themselves networks, in which case this method avoids duplicates in the backup and saves memory space.

        Parameters
        ----------
        module: nn.Module
            Module to save.
        path:
            Path to save the Module.
        """
        os.mkdir(path)
        template = os.path.join(path, "{}")

        args = list(signature(module.__init__).parameters)
        with open(template.format("init.pkl"), "wb") as f:
            pickle.dump((type(module), args), f)

        delegs = []
        for arg in args:
            obj = getattr(module, arg)
            if isinstance(obj, nn.ModuleList):
                obj = list(obj)

            if NeuralNetwork._needs_recursion(obj):
                NeuralNetwork._recursive_save(obj, os.path.join(path, arg))
                delegs.append(arg)
            elif NeuralNetwork._needs_list_recursion(obj):
                newobj = []
                for i, el in enumerate(obj):
                    if NeuralNetwork._needs_recursion(el):
                        NeuralNetwork._recursive_save(
                            el, os.path.join(path, f"{arg}-{i}")
                        )
                        newobj.append(None)
                    else:
                        newobj.append(el)
                NeuralNetwork._save_regular(newobj, template.format(arg))
                delegs.append(arg)
            else:
                NeuralNetwork._save_regular(obj, template.format(arg))

        sd = module.state_dict()
        sd = OrderedDict(
            [
                (key, val)
                for key, val in sd.items()
                if not NeuralNetwork._is_delegated(key, delegs)
            ]
        )

        torch.save(sd, template.format("state_dict.pth"))

    @staticmethod
    def _save_regular(
        obj: object,
        filename: str,
    ) -> None:
        r"""Save object ``obj`` in json or pickle files.

        Parameters
        ----------
        module: nn.Module
            Module to save.
        path:
            Path to save the Module.
        """
        if NeuralNetwork._needs_json(obj):
            with open(f"{filename}.json", "w", encoding="utf-8") as f:
                json.dump(obj, f, ensure_ascii=False, indent=4)
        else:
            with open(f"{filename}.pkl", "wb") as f:
                pickle.dump(obj, f)

    @staticmethod
    def _needs_recursion(obj: object) -> bool:
        r"""Returns ``True`` if ``obj`` is an object that need to be saved recursively, else ``False``.

        Parameters
        ----------
        obj: object
            Any Python object.

        Returns
        -------
        bool
            whether ``obj`` needs to be save recursively.
        """
        return isinstance(obj, NeuralNetwork)

    @staticmethod
    def _needs_list_recursion(obj: object) -> bool:
        r"""Returns ``True`` if ``obj`` is a list whose elements need to be saved recursively, else ``False``.

        Parameters
        ----------
        obj: object
            Any Python object.

        Returns
        -------
        bool
            whether at least one element of ``obj`` needs to be save recursively.
        """
        if not isinstance(obj, list):
            return False
        return any([NeuralNetwork._needs_recursion(el) for el in obj])

    @staticmethod
    def _needs_json(obj: object) -> bool:
        r"""Returns True if the object ``obj`` must be saved in a JSON file.

        Parameters
        ----------
        obj: object
            Any Python object.

        Returns
        -------
        bool
            True if ``obj`` needs to be save in a JSON, else False.
        """
        if obj is None:
            return True
        if isinstance(obj, (bool, int, float, complex, str)):
            return True
        if isinstance(obj, (list, tuple)):
            return all(NeuralNetwork._needs_json(v) for v in obj)
        return False

    @staticmethod
    def _is_delegated(key: str, delegs: List[str]) -> bool:
        r"""Returns True the entry of key ``key`` can be delegated to a recursive save. The attributs that are to be save recursively are contained in ``delegs``.

        Parameters
        ----------
        key: str
            Key of entry.
        delegs: List[str]
            List containing keys of attributes that will be save recursively.

        Returns
        -------
        bool
            True if ``key`` is the name of an attribute whose save can be delegated, else False.
        """
        for prefix in delegs:
            if key.startswith(prefix):
                return True
        return False

    @classmethod
    def load(
        self, module_name: str, module_path: Optional[str] = None
    ) -> "NeuralNetwork":
        r"""Load a network from a local save made using the ``save()`` method.

        Parameters
        ----------
        module_name: str
            Name of the directory in which the model has been saved.
        module_path: str | None
            Path to the previous directory.

        Returns
        -------
        NeuralNetwork
            loaded neural network instance.

        Raises
        ------
        FileNotFoundError
            The specified directory does not exist, or is not a directory.
        """
        if module_path is None:
            path = module_name
        else:
            path = os.path.join(module_path, module_name)

        if not os.path.exists(path):
            raise FileNotFoundError(f"{path} directory does not exist")
        if not os.path.isdir(path):
            raise FileNotFoundError(f"{path} is not a directory")

        net = NeuralNetwork._recursive_load(path)
        net.eval()  # By default in evaluation mode

        return net

    @staticmethod
    def _recursive_load(path: str) -> "NeuralNetwork":
        r"""Make a recursive load of a PyTorch module.

        Parameters
        ----------
        path:
            Path to the module to load.
        """
        template = os.path.join(path, "{}")

        with open(template.format("init.pkl"), "rb") as f:
            type_module, args = pickle.load(f)

        d = {}
        for arg in args:
            if os.path.exists(template.format(f"{arg}.json")):
                with open(template.format(f"{arg}.json"), "r") as f:
                    d.update({arg: json.load(f)})
            elif os.path.exists(template.format(f"{arg}.pkl")):
                with open(template.format(f"{arg}.pkl"), "rb") as f:
                    d.update({arg: pickle.load(f)})
            elif os.path.isdir(template.format(arg)):
                d.update({arg: NeuralNetwork._recursive_load(template.format(arg))})
            else:
                raise RuntimeError("Should never been here.")

            if isinstance(d[arg], list) and any([el is None for el in d[arg]]):
                for i, _ in enumerate(d[arg]):
                    if os.path.isdir(template.format(f"{arg}-{i}")):
                        d[arg][i] = NeuralNetwork._recursive_load(
                            template.format(f"{arg}-{i}")
                        )

        module: nn.Module = type_module(**d)

        sd = module.state_dict()
        sd.update(torch.load(template.format("state_dict.pth")))
        module.load_state_dict(sd)

        return module

    def copy(self) -> "NeuralNetwork":
        r"""Returns a copy of the network. The copy is detached from the original network so any modification of one doesn't modify the other.

        Returns
        -------
        NeuralNetwork
            copy of ``self``.
        """
        d = {
            name: getattr(self, name)
            for name in list(signature(self.__init__).parameters)
        }

        for name in d:
            if NeuralNetwork._needs_recursion(d[name]):
                d[name] = d[name].copy()
            elif NeuralNetwork._needs_list_recursion(d[name]):
                d[name] = [
                    el.copy() if NeuralNetwork._needs_recursion(el) else el
                    for el in d[name]
                ]

        new = type(self)(**d)
        new.load_state_dict(self.state_dict())

        return new

    def __str__(self) -> str:
        d = list(signature(self.__init__).parameters)
        descr = f"{type(self).__name__}:\n"
        for arg in d:
            obj = getattr(self, arg)
            if isinstance(obj, list) and len(obj) > 6:
                obj = obj[:6] + ["..."]
            elif isinstance(obj, tuple):
                obj = obj[:6] + ("...",)
            descr += f"\t{arg}: {obj}\n"
        return descr
